import logging
from logging import getLogger

import torch
import torch.distributed

from chitu.backend import Backend, BackendState
from chitu.executor import Executor
from chitu.global_vars import set_global_variables
from chitu.scheduler import Scheduler
from chitu.task import (
    PackedTasks,
    PackedTasksBase,
    SerializedPackedTasksPayloadType,
    TaskPool,
    TaskType,
    req_encode,
)
from chitu.tensor_parallel import get_tp_group

logger = getLogger(__name__)


def init_logger(logging_level=logging.INFO):
    base_name = __name__.split(".")[0]
    base_logger = getLogger(base_name)
    base_logger.setLevel(logging.INFO)

    def add_rank_to_msg(record):
        if torch.distributed.is_initialized():
            record.msg = f"[Rank {torch.distributed.get_rank()}] {record.msg}"
        return True

    def add_filter_to_all_parent_handlers(cur_logger):
        for handler in cur_logger.handlers:
            handler.addFilter(add_rank_to_msg)
        if cur_logger.parent:
            add_filter_to_all_parent_handlers(cur_logger.parent)

    # If there is no handlers, create a new handler to hold the filter. If not (very likely
    # because we launch from Hydra, and Hydra setup the root logger), a new handler will only
    # duplicate the logs. In this case, we should add the filter to all the existing handlers.
    if base_logger.hasHandlers():  # Including handlers from the parents
        add_filter_to_all_parent_handlers(base_logger)
    else:
        handler = logging.StreamHandler()
        handler.addFilter(add_rank_to_msg)
        base_logger.addHandler(handler)


def chitu_init(args, logging_level=logging.INFO):
    init_logger(logging_level)
    set_global_variables(args)
    Backend.build(args)
    rank = torch.distributed.get_rank()
    if rank == 0:
        scheduler = Scheduler.build(args.scheduler)
        Backend.scheduler = scheduler
    executor = Executor.build(args)
    Backend.executor = executor
    PackedTasks.configure(max_num_tasks=args.infer.max_reqs)


def remove_task_other_device(remove_task_ids):
    # TODO handle in executor?
    if len(remove_task_ids) == 0:
        return
    # Since we are removing, any task type is fine
    task_tensor = PackedTasksBase(
        len(remove_task_ids), remove_task_ids, remove_task_ids, TaskType.Decode
    ).serialize(payload_type=SerializedPackedTasksPayloadType.EndTask, device=0)
    if Backend.args.infer.pp_size > 1:
        torch.distributed.isend(tensor=task_tensor, dst=Backend.args.infer.tp_size)
        if Backend.args.infer.tp_size > 1:
            torch.distributed.broadcast(
                tensor=task_tensor, src=Backend.pp_main_rank, group=get_tp_group()
            )
    elif Backend.args.infer.tp_size > 1:
        torch.distributed.broadcast(tensor=task_tensor, src=0)


def update_ongoing_tasks():
    to_remove = []
    unwait_tasks = []
    logits_list = []
    for ogr in Backend.ongoing_reqs:
        if ogr.handle.is_completed():
            to_remove.append(ogr)
            unwait_tasks.append(ogr.waiting_task)
            logits_list.append(ogr.logits.view(-1, ogr.logits.shape[-1]))
            for task in ogr.waiting_task.tasks:
                task.unwait()
    for tr in to_remove:
        Backend.ongoing_reqs.remove(tr)
    return unwait_tasks, logits_list


def chitu_update(task_ids, rank, world_size):
    if rank == 0:
        TaskPool.display()
    if world_size == 1:
        Backend.scheduler.update(task_ids)
    else:
        unwait_tasks, logits = update_ongoing_tasks()
        for idx, task in enumerate(unwait_tasks):
            Backend.executor.update_response(task, logits[idx])
        unwait_task_ids = [t.task_id for task in unwait_tasks for t in task.tasks]
        removed_decode_task_ids = Backend.scheduler.update(task_ids, unwait_task_ids)
        remove_task_other_device(removed_decode_task_ids)


@torch.inference_mode()
def chitu_run():
    rank = torch.distributed.get_rank()
    world_size = torch.distributed.get_world_size()
    if rank == 0:
        task_ids = Backend.scheduler.schedule()
        if len(task_ids) == 0:  # no tasks to do, but some tasks are waiting
            chitu_update(task_ids, rank, world_size)
            return
        if rank == 0:
            logger.debug(f"Processing {task_ids}")
        tasks = PackedTasks(task_ids, rank)
    else:
        tasks = None
    Backend.executor.step(tasks)

    if Backend.args.infer.pp_size > 1 and rank == 0:
        chitu_update(task_ids, rank, world_size)
    elif rank == 0:
        TaskPool.display()
        removed_decode_task_ids = Backend.scheduler.update(task_ids)
        if world_size != 1:
            remove_task_other_device(removed_decode_task_ids)


def chitu_terminate():
    if torch.distributed.get_rank() == 0:
        Backend.state = BackendState.Terminating
        Backend.executor.step(None)


def chitu_is_terminated():
    return Backend.state == BackendState.Terminated



import os
from dataclasses import dataclass
from logging import getLogger
from typing import Optional, Sequence

import numpy as np
import torch
import torch.distributed

from chitu.backend import Backend, BackendState
from chitu.cache_manager import PagedKVCacheManager
from chitu.global_vars import get_timers
from chitu.task import (
    PackedTasks,
    PackedTasksBase,
    SerializedPackedTasksPayloadType,
    Task,
    TaskType,
    UserRequest,
    req_decode,
    taskid2reqid,
)
from chitu.tensor_parallel import get_tp_group
from chitu.utils import VarLens, top_k_top_p_min_p_sampling_from_probs_torch

logger = getLogger(__name__)

# Although tags are not fully supported in the NCCL backend, they are helpful to understand the code
TASK_TENSOR_TAG = 1
HIDDEN_TENSOR_TAG = 2
LOGIT_TAG = 3


@dataclass
class OngoingRequests:
    waiting_task: PackedTasks
    handle: torch.distributed.distributed_c10d.Work
    logits: torch.Tensor


class Executor:
    @staticmethod
    def build(args):
        if args.infer.pp_size > 1:
            return PipeTensorExecutor(args)
        elif args.infer.tp_size > 1:
            return TensorExecutor(args)
        else:
            return NormalExecutor(args)

    def __init__(self, args):
        self.timers = get_timers()
        pass

    def step(
        self,
        tasks: PackedTasksBase,
    ):
        pass

    def _prepare_seq_lens_for_decode(self, tasks):
        seq_lens = []
        for req_id in tasks.req_ids:
            seq_len = Backend.cache_manager.seq_lens[req_id]
            seq_lens.append(seq_len)
        return seq_lens

    def _prepare_new_tokens_for_decode(self, tasks):
        new_tokens = []
        for task in tasks.tasks:
            new_tokens.append(task.next_token)
        new_tokens = torch.tensor(
            new_tokens, device="cuda", dtype=torch.long
        ).unsqueeze(1)
        return new_tokens


class NormalExecutor(Executor):

    def __init__(self, args):
        super().__init__(args)

    def update_response(self, tasks: PackedTasks, logits: torch.Tensor):
        logits = logits.view(-1, logits.shape[-1])
        assert (
            len(tasks.tasks) == logits.shape[0]
        ), f"logtis has shape {logits.shape}, but there are {len(tasks.tasks)} tasks"
        for it, task in enumerate(tasks.tasks):
            if (
                task.req.params.frequency_penalty > 0
                and task.task_type == TaskType.Decode
                and len(task.response) > 0
            ):
                logits[it].index_add_(
                    -1,
                    task.response.to_tensor(),
                    -task.req.params.frequency_penalty
                    * torch.ones(
                        (len(task.response),),
                        dtype=logits.dtype,
                        device=logits.device,
                    ),
                )
        if tasks.is_all_greedy:
            tokens = torch.argmax(logits, dim=-1)
        else:
            probs = torch.softmax(logits / tasks.temperatures.view(-1, 1), dim=-1)
            tokens = top_k_top_p_min_p_sampling_from_probs_torch(
                probs, tasks.top_ks, tasks.top_ps
            )
        tokens_cpu = tokens.cpu()
        for it, task in enumerate(tasks.tasks):
            task.update_response(tokens_cpu[it].item(), tokens[it], logits[it])

    def propagate_tasks(self, tasks: Optional[PackedTasksBase]):
        """Make every ranks know the task metadata"""
        return tasks  # Need to do nothing if not parallelized

    def prefill_step(self, tasks: PackedTasksBase):
        logger.debug(f"Prefill step: {tasks.task_ids}")
        varlens = VarLens(tasks.tokens, "cuda")
        self.timers("prefill").start()
        Backend.cache_manager.curr_varlens = varlens
        Backend.cache_manager.curr_req_ids = tasks.req_ids
        logits = Backend.model.prefill(tasks.tokens)
        self.timers("prefill").stop()
        self.update_response(tasks, logits)
        for it in range(tasks.num_tasks):
            tasks.tasks[it].start_decoding()
        varlens = VarLens(tasks.tokens, "cuda")
        Backend.cache_manager.finalize_cache_all_prefill(tasks.req_ids, varlens)
        return logits

    def decode_step(self, tasks: PackedTasksBase):
        logger.debug(f"Decode step: {tasks.task_ids}")
        Backend.cache_manager.prepare_cache_decode(tasks.req_ids)
        Backend.cache_manager.curr_req_ids = tasks.req_ids
        if isinstance(Backend.cache_manager, PagedKVCacheManager):
            Backend.cache_manager.prepare_block_table_for_decode(tasks.req_ids)
        self.timers("decode").start()
        new_tokens = self._prepare_new_tokens_for_decode(tasks)
        seq_lens = self._prepare_seq_lens_for_decode(tasks)
        self.timers("decode-model").start()
        logits = Backend.model.decode(new_tokens, seq_lens)
        self.timers("decode-model").stop()
        self.timers("decode").stop()
        self.update_response(tasks, logits)
        Backend.cache_manager.finalize_cache_single_decode(tasks.req_ids)
        return logits

    def step(
        self,
        tasks: PackedTasksBase,
    ):
        tasks = self.propagate_tasks(tasks)
        if tasks is None:
            return
        if tasks.task_type == TaskType.Prefill:
            return self.prefill_step(tasks)
        elif tasks.task_type == TaskType.Decode:
            return self.decode_step(tasks)
        else:
            raise NotImplementedError  # Hybrid task not implemented


class PipeTensorExecutor(NormalExecutor):

    def __init__(self, args):
        super().__init__(args)
        self.rank = torch.distributed.get_rank()
        self.local_rank = int(os.environ.get("LOCAL_RANK", 0))
        self.tp_size = args.infer.tp_size
        self.pp_size = args.infer.pp_size
        self.pp_stage = Backend.pp_stage
        self.pp_main_rank = Backend.pp_main_rank
        self.pp_end_stage = Backend.pp_end_stage
        self.last_pp_main_rank = self.pp_end_stage * self.tp_size
        self.tp_group = get_tp_group()

    def prefill_step(self, tasks: PackedTasksBase):
        varlens = VarLens(tasks.tokens, device=self.local_rank)
        Backend.cache_manager.curr_varlens = varlens
        Backend.cache_manager.curr_req_ids = tasks.req_ids

        if self.rank == 0:
            inp = torch.from_numpy(np.concatenate(tasks.tokens)).to(self.local_rank)
            if self.tp_size > 1:
                torch.distributed.broadcast(
                    tensor=inp, src=self.pp_main_rank, group=self.tp_group
                )  # TODO main rank 处理统一放后面
        else:
            if self.pp_stage == 0:
                inp = torch.empty(
                    [sum(len(seq) for seq in tasks.tokens)],
                    dtype=torch.int64,
                    device=self.local_rank,
                )
            else:
                inp = torch.empty(
                    [
                        sum([len(token) for token in tasks.tokens]),
                        Backend.model.params.dim,
                    ],
                    device=self.local_rank,
                )
            if self.rank == self.pp_main_rank:
                torch.distributed.recv(
                    tensor=inp, src=self.rank - self.tp_size, tag=HIDDEN_TENSOR_TAG
                )
            if self.tp_size > 1:
                torch.distributed.broadcast(
                    tensor=inp, src=self.pp_main_rank, group=self.tp_group
                )
        self.timers("prefill").start()

        out = Backend.model.prefill(inp)

        self.timers("prefill").stop()
        if self.rank == self.pp_main_rank and self.pp_stage != self.pp_end_stage:
            torch.distributed.isend(
                tensor=out.contiguous(),  # contiguous() is necessary for NCCL
                dst=self.rank + self.tp_size,
                tag=HIDDEN_TENSOR_TAG,
            )
        elif self.rank == self.pp_main_rank and self.pp_stage == self.pp_end_stage:
            # send logits to rank 0 to get response words
            torch.cuda.synchronize(self.local_rank)
            torch.distributed.isend(
                tensor=out.contiguous(),  # contiguous() is necessary for NCCL
                dst=0,
                tag=LOGIT_TAG,
            )
        Backend.cache_manager.finalize_cache_all_prefill(tasks.req_ids, varlens)
        return out

    def decode_step(self, tasks: PackedTasksBase):
        Backend.cache_manager.prepare_cache_decode(tasks.req_ids)
        Backend.cache_manager.curr_req_ids = tasks.req_ids
        if isinstance(Backend.cache_manager, PagedKVCacheManager):
            Backend.cache_manager.prepare_block_table_for_decode(tasks.req_ids)
        if self.rank == 0:
            inp = self._prepare_new_tokens_for_decode(tasks)  # tensor
            if self.tp_size > 1:
                torch.distributed.broadcast(
                    tensor=inp, src=self.pp_main_rank, group=self.tp_group
                )
        else:
            if self.pp_stage == 0:
                inp = torch.empty(
                    [tasks.num_tasks, 1], dtype=torch.int64, device=self.local_rank
                )
            else:
                inp = torch.empty(
                    [tasks.num_tasks, 1, Backend.model.params.dim],
                    device=self.local_rank,
                )
            if self.rank == self.pp_main_rank:
                torch.distributed.recv(
                    tensor=inp, src=self.rank - self.tp_size, tag=HIDDEN_TENSOR_TAG
                )
            if self.tp_size > 1:
                torch.distributed.broadcast(
                    tensor=inp, src=self.pp_main_rank, group=self.tp_group
                )
        self.timers("decode").start()
        seq_lens = self._prepare_seq_lens_for_decode(tasks)
        self.timers("decode-model").start()
        out = Backend.model.decode(inp, seq_lens)
        self.timers("decode-model").stop()
        self.timers("decode").stop()
        if self.rank == self.pp_main_rank and self.pp_stage != self.pp_end_stage:
            torch.distributed.isend(
                tensor=out.contiguous(),  # contiguous() is necessary for NCCL
                dst=self.rank + self.tp_size,
                tag=HIDDEN_TENSOR_TAG,
            )
        elif self.rank == self.pp_main_rank and self.pp_stage == self.pp_end_stage:
            # Send logits to rank 0 to get response words
            out = out.view(out.shape[0], -1)
            torch.distributed.isend(
                out.contiguous(),  # contiguous() is necessary for NCCL
                dst=0,
                tag=LOGIT_TAG,
            )
        Backend.cache_manager.finalize_cache_single_decode(tasks.req_ids)
        return out

    def _recv_logits(self, tasks):
        logits = torch.empty(
            [tasks.num_tasks, Backend.model.vocab_size],
            device=self.local_rank,
            dtype=torch.float,
        )
        handle = torch.distributed.irecv(
            logits, src=self.last_pp_main_rank, tag=LOGIT_TAG
        )
        Backend.ongoing_reqs.append(OngoingRequests(tasks, handle, logits))
        for it, task in enumerate(tasks.tasks):
            task.wait(handle)

    def propagate_tasks(self, tasks: Optional[PackedTasksBase]):
        remove_kvcache = False

        # PP stage 0 initialzie from the argument. PP stage >= 1 recv task tensor from stage - 1
        if self.rank == 0:
            if Backend.state == BackendState.Running:
                task_tensor = tasks.serialize(device=self.local_rank)
            else:
                assert Backend.state == BackendState.Terminating
                task_tensor = PackedTasksBase.serialize_special(
                    SerializedPackedTasksPayloadType.TerminateBackend,
                    device=self.local_rank,
                )
            if self.tp_size > 1:
                torch.distributed.broadcast(
                    tensor=task_tensor, src=self.pp_main_rank, group=self.tp_group
                )

        else:
            task_tensor = PackedTasksBase.empty_serialization(device=self.local_rank)
            if self.rank == self.pp_main_rank:
                torch.distributed.recv(
                    tensor=task_tensor,
                    src=self.rank - self.tp_size,
                    tag=TASK_TENSOR_TAG,
                )
            if self.tp_size > 1:
                torch.distributed.broadcast(
                    tensor=task_tensor, src=self.pp_main_rank, group=self.tp_group
                )
            task_tensor_type, tasks = PackedTasksBase.deserialize(task_tensor)
            if task_tensor_type == SerializedPackedTasksPayloadType.TerminateBackend:
                Backend.state = BackendState.Terminating
            if task_tensor_type == SerializedPackedTasksPayloadType.EndTask:
                remove_kvcache = True

        if self.rank == self.pp_main_rank and self.pp_stage != self.pp_end_stage:
            torch.distributed.isend(
                tensor=task_tensor, dst=self.rank + self.tp_size, tag=TASK_TENSOR_TAG
            )

        if Backend.state == BackendState.Terminating:
            Backend.state = BackendState.Terminated
        if Backend.state == BackendState.Terminated:
            return None

        if remove_kvcache:
            for rid in tasks.req_ids:
                Backend.cache_manager.finalize_cache_all_decode(rid)
            return None

        return tasks

    def step(
        self,
        tasks: Optional[PackedTasksBase] = None,
    ):
        # Run tasks
        super().step(tasks)

        if Backend.state == BackendState.Terminated:
            return
        # Rank 0 recv final logits from last PP stage
        if self.rank == 0:
            if tasks.task_type == TaskType.Prefill:
                # After prefill, new decode tasks are created
                for it in range(tasks.num_tasks):
                    tasks.tasks[it].start_decoding()
                    tasks.tasks[it].wait(None)
            self._recv_logits(tasks)


class TensorExecutor(NormalExecutor):
    def __init__(self, args):
        super().__init__(args)
        self.rank = torch.distributed.get_rank()
        self.local_rank = int(os.environ.get("LOCAL_RANK", 0))

    def propagate_tasks(self, tasks: Optional[PackedTasksBase]):
        """Broadcast task metadata from rank 0 to all other ranks"""
        remove_kvcache = False
        if Backend.state == BackendState.Running:
            task_tensor = (
                tasks.serialize(device=self.local_rank)
                if self.rank == 0
                else PackedTasksBase.empty_serialization(device=self.local_rank)
            )
        else:
            assert Backend.state == BackendState.Terminating
            task_tensor = PackedTasksBase.serialize_special(
                SerializedPackedTasksPayloadType.TerminateBackend,
                device=self.local_rank,
            )
            Backend.state = BackendState.Terminated
        torch.distributed.broadcast(tensor=task_tensor, src=0)

        if self.rank != 0:
            task_tensor_type, tasks = PackedTasksBase.deserialize(task_tensor)

        if self.rank != 0:
            if task_tensor_type == SerializedPackedTasksPayloadType.TerminateBackend:
                Backend.state = BackendState.Terminated
            if task_tensor_type == SerializedPackedTasksPayloadType.EndTask:
                remove_kvcache = True
        if Backend.state == BackendState.Terminated:
            return None

        if remove_kvcache:
            for rid in tasks.req_ids:
                Backend.cache_manager.finalize_cache_all_decode(rid)
            return None

        return tasks

    def prefill_step(self, tasks):
        if self.rank == 0:
            tokens = torch.from_numpy(np.concatenate(tasks.tokens)).to(self.rank)
        else:
            tokens = torch.empty(
                [sum(len(seq) for seq in tasks.tokens)],
                dtype=torch.int64,
                device=self.local_rank,
            )
        torch.distributed.broadcast(tensor=tokens, src=0)

        varlens = VarLens(tasks.tokens, device=self.local_rank)
        self.timers("prefill").start()
        Backend.cache_manager.curr_varlens = varlens
        Backend.cache_manager.curr_req_ids = tasks.req_ids
        logits = Backend.model.prefill(tokens, varlens=varlens)
        self.timers("prefill").stop()
        if self.rank == 0:
            self.update_response(tasks, logits)
            for it in range(tasks.num_tasks):
                tasks.tasks[it].start_decoding()
        Backend.cache_manager.finalize_cache_all_prefill(tasks.req_ids, varlens)
        return logits

    def decode_step(self, tasks):
        if self.rank == 0:
            tokens = []
            for task in tasks.tasks:
                tokens.append(task.next_token)
            tokens = torch.tensor(
                tokens, dtype=torch.int64, device=self.local_rank
            ).unsqueeze(1)
            # TODO
            # inp = self._prepare_new_tokens_for_decode(tasks)
        else:
            tokens = torch.empty(
                [tasks.num_tasks, 1], dtype=torch.int64, device=self.local_rank
            )
        torch.distributed.broadcast(tensor=tokens, src=0)

        Backend.cache_manager.prepare_cache_decode(tasks.req_ids)
        Backend.cache_manager.curr_req_ids = tasks.req_ids
        if isinstance(Backend.cache_manager, PagedKVCacheManager):
            Backend.cache_manager.prepare_block_table_for_decode(tasks.req_ids)
        self.timers("decode").start()
        seq_lens = self._prepare_seq_lens_for_decode(tasks)
        self.timers("decode-model").start()
        logits = Backend.model.decode(tokens, seq_lens)
        self.timers("decode-model").stop()
        self.timers("decode").stop()
        if self.rank == 0:
            self.update_response(tasks, logits)
        Backend.cache_manager.finalize_cache_single_decode(tasks.req_ids)
        return logits



from functools import partial
from typing import Tuple

import torch
import torch.nn as nn
from torch import nn

from chitu.global_vars import get_timers
from chitu.muxi_utils import tbsgemm
from chitu.tensor_parallel import ColumnParallelLinear, RowParallelLinear


class NormAndQuant(torch.nn.Module):
    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def _norm(self, x):
        return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)

    def forward(self, x):
        # output = self._norm(x.float()).type_as(x)
        # return output * self.weight

        output, scale = tbsgemm.normAndQuant(x, self.weight)
        return (output, scale)

    @staticmethod
    def from_float(module, model_arch_only=False):
        new_module = NormAndQuant(module.weight.numel())
        if not model_arch_only:
            new_module.weight.data = module.weight.data
        return new_module


@torch.no_grad()
def quant_act(act):
    act_shape = act.shape
    act.view(-1, act_shape[-1])
    scales = act.abs().max(dim=-1, keepdim=True)[0]
    scales = scales.to(torch.float)
    scales.clamp_(min=1e-5).div_(127.0)
    aa = act.div(scales).round_()
    return aa.to(torch.int8).view(-1, act_shape[-1]), scales.view(-1)


@torch.no_grad()
def quant_weight(w):
    scales = w.abs().max(dim=-1, keepdim=True)[0]
    scales = scales.to(torch.float)
    scales.clamp_(min=1e-5).div_(127.0)
    ww = w.div(scales).round_()
    return ww.to(torch.int8), scales.view(-1)


class W8A8Linear(nn.Module):
    def __init__(
        self,
        in_features,
        out_features,
        bias=True,
        quantize_output=False,
        pre_norm=None,
    ):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features

        self.pre_norm = pre_norm

        self.register_buffer(
            "weight",
            torch.zeros(
                self.out_features,
                self.in_features,
                dtype=torch.int8,
                requires_grad=False,
            ),
        )
        self.register_buffer(
            "org_weight",
            torch.randn(
                self.out_features,
                self.in_features,
                dtype=torch.float16,
                requires_grad=False,
            ),
        )
        self.register_buffer(
            "scale_channel",
            torch.ones(
                [self.out_features],
                dtype=torch.float,
                requires_grad=False,
            ),
        )
        if bias:
            self.register_buffer(
                "bias",
                torch.zeros(
                    (self.out_features,), dtype=torch.float16, requires_grad=False
                ),
            )
        else:
            self.register_buffer("bias", None)

        self.act_quant_name = "per_token"
        self.act_quant = quant_act
        if quantize_output:
            self.output_quant_name = self.act_quant_name
            self.output_quant = self.act_quant
        else:
            self.output_quant_name = "None"
            self.output_quant = lambda x: x

        self.timers = get_timers()

    def to(self, *args, **kwargs):
        super(W8A8Linear, self).to(*args, **kwargs)
        self.weight = self.weight.to(*args, **kwargs)
        if self.bias is not None:
            self.bias = self.bias.to(*args, **kwargs)
        return self

    @torch.no_grad()
    def forward(self, x):

        if isinstance(x, Tuple):
            q_x = x[0]
            act_scale = x[1]
            if q_x.dim() == 2:
                out = tbsgemm.mm(
                    self.weight, q_x, self.scale_channel, act_scale.to(torch.float32)
                )
            else:
                bs, seq, _ = q_x.shape
                q_x = q_x.view(bs * seq, q_x.shape[-1])
                out = tbsgemm.mm(
                    self.weight, q_x, self.scale_channel, act_scale.to(torch.float32)
                )
                out = out.reshape(bs, seq, -1)
            if self.bias is not None:
                out += self.bias
            return out
        else:
            if x.dim() == 2:
                m, _ = x.shape
                # self.timers("quanting").start()
                # q_x, act_scale = self.act_quant(x)
                q_x, act_scale = tbsgemm.quant(x)
                # self.timers("quanting").stop()
                # self.timers("w8a8gemm").start()
                out = tbsgemm.mm(
                    self.weight, q_x, self.scale_channel, act_scale.to(torch.float32)
                )
                # self.timers("w8a8gemm").stop()
            else:

                bs, seq, _ = x.shape
                x = x.view(bs * seq, x.shape[-1])
                # self.timers("quanting").start()
                # q_x, act_scale = self.act_quant(x)
                q_x, act_scale = tbsgemm.quant(x)
                # self.timers("quanting").stop()
                # self.timers("w8a8gemv").start()
                out = tbsgemm.mm(
                    self.weight, q_x, self.scale_channel, act_scale.to(torch.float32)
                )
                # self.timers("w8a8gemv").stop()
                out = out.reshape(bs, seq, -1)

            if self.bias is not None:
                out += self.bias

            return out

    @staticmethod
    def from_float(
        module,
        weight_quant="per_channel",
        act_quant="per_token",
        quantize_output=False,
        model_arch_only=False,
    ):
        assert isinstance(
            module, (torch.nn.Linear, ColumnParallelLinear, RowParallelLinear)
        )
        new_module = W8A8Linear(
            module.in_features,
            module.out_features,
            module.bias is not None,
            quantize_output=quantize_output,
        )
        if not model_arch_only:
            # new_module.weight = quantize_weight_per_channel_absmax(
            #    module.weight, n_bits=8
            # )
            new_module.org_weight = module.weight
            ww, scl = quant_weight(module.weight)
            new_module.weight = ww
            new_module.scale_channel = scl
            if module.bias is not None:
                new_module.bias = module.bias
        return new_module

    def __repr__(self):
        return f"W8A8Linear({self.in_features}, {self.out_features}, bias={self.bias is not None})"



"""
This file has adaption of open-source code from the following sources:
- The triton implementation of MLA is credited to monellz (https://github.com/monellz/vllm/commit/feebaa7c063be6bfb590a876741aeef1c5f58cf8)
"""

import triton
import triton.language as tl


_mla_attn_kernel_configs = [
    triton.Config(
        {"BLOCK_N": block_n},
        num_stages=num_stages,
        num_warps=4,
    )
    for block_n in [32, 64]
    for num_stages in [1, 2]
]


@triton.autotune(configs=_mla_attn_kernel_configs, key=[])
@triton.jit
def _mla_attn_kernel(
    Q_nope,
    Q_pe,
    Kv_c_cache,
    K_pe_cache,
    Req_to_tokens,
    B_seq_len,
    O,
    sm_scale,
    stride_q_nope_bs,
    stride_q_nope_h,
    stride_q_pe_bs,
    stride_q_pe_h,
    stride_kv_c_bs,
    stride_k_pe_bs,
    stride_req_to_tokens_bs,
    stride_o_b,
    stride_o_h,
    stride_o_s,
    BLOCK_H: tl.constexpr,
    BLOCK_N: tl.constexpr,
    NUM_KV_SPLITS: tl.constexpr,
    PAGE_SIZE: tl.constexpr,
    HEAD_DIM_CKV: tl.constexpr,
    HEAD_DIM_KPE: tl.constexpr,
):
    cur_batch = tl.program_id(0)
    cur_head_id = tl.program_id(1)
    split_kv_id = tl.program_id(2)

    cur_batch_seq_len = tl.load(B_seq_len + cur_batch)

    offs_d_ckv = tl.arange(0, HEAD_DIM_CKV)
    cur_head = cur_head_id * BLOCK_H + tl.arange(0, BLOCK_H)
    offs_q_nope = (
        cur_batch * stride_q_nope_bs
        + cur_head[:, None] * stride_q_nope_h
        + offs_d_ckv[None, :]
    )
    q_nope = tl.load(Q_nope + offs_q_nope)

    offs_d_kpe = tl.arange(0, HEAD_DIM_KPE)
    offs_q_pe = (
        cur_batch * stride_q_pe_bs
        + cur_head[:, None] * stride_q_pe_h
        + offs_d_kpe[None, :]
    )
    q_pe = tl.load(Q_pe + offs_q_pe)

    e_max = tl.zeros([BLOCK_H], dtype=tl.float32) - float("inf")
    e_sum = tl.zeros([BLOCK_H], dtype=tl.float32)
    acc = tl.zeros([BLOCK_H, HEAD_DIM_CKV], dtype=tl.float32)

    kv_len_per_split = tl.cdiv(cur_batch_seq_len, NUM_KV_SPLITS)
    split_kv_start = kv_len_per_split * split_kv_id
    split_kv_end = tl.minimum(split_kv_start + kv_len_per_split, cur_batch_seq_len)

    if split_kv_end > split_kv_start:
        for start_n in range(split_kv_start, split_kv_end, BLOCK_N):
            offs_n = start_n + tl.arange(0, BLOCK_N)
            kv_page_number = tl.load(
                Req_to_tokens
                + stride_req_to_tokens_bs * cur_batch
                + offs_n // PAGE_SIZE,
                mask=offs_n < split_kv_end,
                other=0,
            )
            kv_loc = kv_page_number * PAGE_SIZE + offs_n % PAGE_SIZE
            offs_k_c = kv_loc[None, :] * stride_kv_c_bs + offs_d_ckv[:, None]
            k_c = tl.load(
                Kv_c_cache + offs_k_c, mask=offs_n[None, :] < split_kv_end, other=0.0
            )

            qk = tl.dot(q_nope, k_c.to(q_nope.dtype))

            offs_k_pe = kv_loc[None, :] * stride_k_pe_bs + offs_d_kpe[:, None]
            k_pe = tl.load(
                K_pe_cache + offs_k_pe, mask=offs_n[None, :] < split_kv_end, other=0.0
            )

            qk += tl.dot(q_pe, k_pe.to(q_pe.dtype))
            qk *= sm_scale

            qk = tl.where(offs_n[None, :] < split_kv_end, qk, float("-inf"))

            v_c = tl.trans(k_c)

            n_e_max = tl.maximum(tl.max(qk, 1), e_max)
            re_scale = tl.exp(e_max - n_e_max)
            p = tl.exp(qk - n_e_max[:, None])
            acc *= re_scale[:, None]
            acc += tl.dot(p.to(v_c.dtype), v_c)

            e_sum = e_sum * re_scale + tl.sum(p, 1)
            e_max = n_e_max
        offs_o = (
            cur_batch * stride_o_b
            + cur_head[:, None] * stride_o_h
            + split_kv_id * stride_o_s
            + offs_d_ckv[None, :]
        )
        tl.store(O + offs_o, acc / e_sum[:, None])
        offs_o_1 = (
            cur_batch * stride_o_b
            + cur_head * stride_o_h
            + split_kv_id * stride_o_s
            + HEAD_DIM_CKV
        )
        tl.store(O + offs_o_1, e_max + tl.log(e_sum))


def _mla_attn(
    q_nope,
    q_pe,
    kv_c_cache,
    k_pe_cache,
    attn_logits,
    req_to_tokens,
    b_seq_len,
    num_kv_splits,
    sm_scale,
    page_size,
):
    batch_size, head_num = q_nope.shape[0], q_nope.shape[1]
    head_dim_ckv = q_nope.shape[-1]
    head_dim_kpe = q_pe.shape[-1]

    BLOCK_H = 16
    grid = (
        batch_size,
        triton.cdiv(head_num, BLOCK_H),
        num_kv_splits,
    )

    _mla_attn_kernel[grid](
        q_nope,
        q_pe,
        kv_c_cache,
        k_pe_cache,
        req_to_tokens,
        b_seq_len,
        attn_logits,
        sm_scale,
        q_nope.stride(0),
        q_nope.stride(1),
        q_pe.stride(0),
        q_pe.stride(1),
        kv_c_cache.stride(-2),
        k_pe_cache.stride(-2),
        req_to_tokens.stride(0),
        attn_logits.stride(0),
        attn_logits.stride(1),
        attn_logits.stride(2),
        BLOCK_H=BLOCK_H,
        NUM_KV_SPLITS=num_kv_splits,
        PAGE_SIZE=page_size,
        HEAD_DIM_CKV=head_dim_ckv,
        HEAD_DIM_KPE=head_dim_kpe,
    )


@triton.jit
def _mla_softmax_reducev_kernel(
    Logits,
    B_seq_len,
    O,
    stride_l_b,
    stride_l_h,
    stride_l_s,
    stride_o_b,
    stride_o_h,
    NUM_KV_SPLITS: tl.constexpr,
    HEAD_DIM_CKV: tl.constexpr,
):
    cur_batch = tl.program_id(0)
    cur_head = tl.program_id(1)
    cur_batch_seq_len = tl.load(B_seq_len + cur_batch)

    offs_d_ckv = tl.arange(0, HEAD_DIM_CKV)

    e_sum = 0.0
    e_max = -float("inf")
    acc = tl.zeros([HEAD_DIM_CKV], dtype=tl.float32)

    offs_l = cur_batch * stride_l_b + cur_head * stride_l_h + offs_d_ckv
    offs_l_1 = cur_batch * stride_l_b + cur_head * stride_l_h + HEAD_DIM_CKV

    for split_kv_id in range(0, NUM_KV_SPLITS):
        kv_len_per_split = tl.cdiv(cur_batch_seq_len, NUM_KV_SPLITS)
        split_kv_start = kv_len_per_split * split_kv_id
        split_kv_end = tl.minimum(split_kv_start + kv_len_per_split, cur_batch_seq_len)

        if split_kv_end > split_kv_start:
            logits = tl.load(Logits + offs_l + split_kv_id * stride_l_s)
            logits_1 = tl.load(Logits + offs_l_1 + split_kv_id * stride_l_s)

            n_e_max = tl.maximum(logits_1, e_max)
            old_scale = tl.exp(e_max - n_e_max)
            acc *= old_scale
            exp_logic = tl.exp(logits_1 - n_e_max)
            acc += exp_logic * logits

            e_sum = e_sum * old_scale + exp_logic
            e_max = n_e_max

    tl.store(
        O + cur_batch * stride_o_b + cur_head * stride_o_h + offs_d_ckv,
        acc / e_sum,
    )


def _mla_softmax_reducev(
    logits,
    o,
    b_seq_len,
    num_kv_splits,
):
    batch_size, head_num, head_dim_ckv = o.shape[0], o.shape[1], o.shape[2]
    grid = (batch_size, head_num)
    _mla_softmax_reducev_kernel[grid](
        logits,
        b_seq_len,
        o,
        logits.stride(0),
        logits.stride(1),
        logits.stride(2),
        o.stride(0),
        o.stride(1),
        NUM_KV_SPLITS=num_kv_splits,
        HEAD_DIM_CKV=head_dim_ckv,
        num_warps=4,
        num_stages=2,
    )


def mla_decode(
    q_nope,
    q_pe,
    kv_c_cache,
    k_pe_cache,
    o,
    req_to_tokens,
    b_seq_len,
    attn_logits,
    num_kv_splits,
    sm_scale,
    page_size,
):
    assert num_kv_splits == attn_logits.shape[2]
    _mla_attn(
        q_nope,
        q_pe,
        kv_c_cache,
        k_pe_cache,
        attn_logits,
        req_to_tokens,
        b_seq_len,
        num_kv_splits,
        sm_scale,
        page_size,
    )
    _mla_softmax_reducev(
        attn_logits,
        o,
        b_seq_len,
        num_kv_splits,
    )



import functools
from typing import Any, List, Mapping

import torch
from torch import nn

from chitu.attn_backend import AttnBackend
from chitu.models.model_hf_llama import (
    FeedForwardHFLlama,
    TransformerBlockHFLlama,
    TransformerHFLlama,
)
from chitu.tensor_parallel import ColumnParallelLinear, RowParallelLinear


class FeedForwardExpertHFMixtral(FeedForwardHFLlama):
    def __init__(
        self, dim: int, hidden_dim: int, op_impl: str, merge_gate_up: bool = True
    ):
        super().__init__(dim, hidden_dim, op_impl=op_impl, merge_gate_up=merge_gate_up)


class SparseMoeBlockHFMixtral(nn.Module):
    def __init__(
        self,
        dim: int,
        hidden_dim: int,
        num_experts: int,
        top_k: int,
        op_impl: str,
        merge_gate_up: bool = True,
    ):
        super().__init__()
        self.num_experts = num_experts
        self.top_k = top_k

        # num_experts is very low, so don't use ColumnParallelLinear
        self.gate = RowParallelLinear(
            dim, num_experts, has_bias=False, input_is_parallel=False
        )

        self.experts = nn.ModuleList(
            [
                FeedForwardExpertHFMixtral(
                    dim, hidden_dim, op_impl=op_impl, merge_gate_up=merge_gate_up
                )
                for _ in range(num_experts)
            ]
        )

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        input_shape = hidden_states.shape
        hidden_dim = hidden_states.shape[-1]
        hidden_states = hidden_states.view(-1, hidden_dim)
        router_logits = self.gate(hidden_states)  # (batch * sequence_length, n_experts)

        routing_weights = torch.nn.functional.softmax(
            router_logits, dim=-1, dtype=torch.float
        )
        routing_weights, selected_experts = torch.topk(
            routing_weights, self.top_k, dim=-1
        )
        routing_weights /= routing_weights.sum(dim=-1, keepdim=True)
        routing_weights = routing_weights.to(hidden_states.dtype)

        final_hidden_states = torch.zeros_like(hidden_states)

        # One hot encode the selected experts to create an expert mask
        # this will be used to easily index which expert is going to be sollicitated
        expert_mask = torch.nn.functional.one_hot(
            selected_experts, num_classes=self.num_experts
        ).permute(
            2, 0, 1
        )  # (n_experts, batch * sequence_length, top_k)

        # Loop over all available experts in the model and perform the computation on each expert
        for expert_idx in range(self.num_experts):
            expert_layer = self.experts[expert_idx]
            batch_seq_idx, expert_idx = torch.where(expert_mask[expert_idx])

            # Index the correct hidden states and compute the expert hidden state for
            # the current expert. We need to make sure to multiply the output hidden
            # states by `routing_weights` on the corresponding tokens (top-1 and top-2)
            current_state = hidden_states[batch_seq_idx].reshape(-1, hidden_dim)
            current_hidden_states = (
                expert_layer(current_state)
                * routing_weights[batch_seq_idx, expert_idx, None]
            )

            # However `index_add_` only support torch tensors for indexing so we'll use
            # the `batch_seq_idx` tensor here.
            final_hidden_states.index_add_(
                0, batch_seq_idx, current_hidden_states.to(hidden_states.dtype)
            )
        final_hidden_states = final_hidden_states.reshape(input_shape)
        return final_hidden_states


class TransformerBlockHFMixtral(TransformerBlockHFLlama):
    def __init__(
        self,
        layer_id: int,
        args,
        cache,
        attn_backend,
        op_impl="torch",
        rotary_type="hf-llama",
        mlp_type=SparseMoeBlockHFMixtral,
        merge_qkv_gate_up=True,
    ):
        super().__init__(
            layer_id,
            args,
            cache,
            attn_backend=attn_backend,
            op_impl=op_impl,
            rotary_type=rotary_type,
            mlp_type=functools.partial(
                mlp_type,
                num_experts=args.num_local_experts,
                top_k=args.num_experts_per_tok,
            ),
            merge_qkv_gate_up=merge_qkv_gate_up,
        )


class TransformerHFMixtral(TransformerHFLlama):
    def __init__(
        self,
        params,
        cache,
        *,
        max_position_embeddings: int,
        pipeline_parallel_size: int,
        model_parallel_size: int,
        attn_backend: AttnBackend,
        rotary_type: str = "hf-llama",
        layer_type: type = TransformerBlockHFMixtral,
        merge_qkv_gate_up: bool = True,
        op_impl: str = "torch",
        **kvargs,
    ):
        super().__init__(
            params,
            cache,
            max_position_embeddings=max_position_embeddings,
            pipeline_parallel_size=pipeline_parallel_size,
            model_parallel_size=model_parallel_size,
            attn_backend=attn_backend,
            rotary_type=rotary_type,
            layer_type=layer_type,
            merge_qkv_gate_up=merge_qkv_gate_up,
            op_impl=op_impl,
            **kvargs,
        )

    def _get_tensor_row_parallel_layer_names(self) -> List[str]:
        return super()._get_tensor_row_parallel_layer_names() + [
            "gate",  # MoE gate
        ]

    def load_state_dict_parallel(
        self,
        state_dict: Mapping[str, Any],
        skip_preprocess: bool = False,
        *args,
        **kwargs,
    ):
        if not skip_preprocess:

            def map_mixtral_key(k):
                k = k.replace(".block_sparse_moe.", ".mlp.")
                k = k.replace(".w1.", ".gate_proj.")
                k = k.replace(".w3.", ".up_proj.")
                k = k.replace(".w2.", ".down_proj.")
                return k

            state_dict = {map_mixtral_key(k): v for k, v in state_dict.items()}

        super().load_state_dict_parallel(
            state_dict, skip_preprocess=skip_preprocess, *args, **kwargs
        )



import struct
from typing import Tuple

import torch
import triton
import triton.language as tl

from chitu.triton_kernels import *


def auto_retry_triton_compilation(fn):
    """
    Avoid file confict introduced by Triton compiler.

    Triton kernels needs to be compiled at the first run, and the Triton compiler uses
    `~/.triton/cache/` for temporary files. However, in distributed envrionment where
    `~` is mounted by NFS, these files may conflict due to the lack of locking mechanism
    in NFS.

    This function simply retries the compilation if the error is related to file conflict.
    """

    # TODO: Use a better way to avoid file conflict. For example, we can create a symlink
    # from `~/.triton/cache` to a local directory, or we can make use of `torch.distributed`
    # to synchronize the compilation.

    import random
    import time

    def wrapped(*args, **kwargs):
        i = 0
        while True:
            try:
                return fn(*args, **kwargs)
            except Exception as e:
                if i >= 30:
                    raise e
                i += 1
                msg = str(e)
                if (
                    "cannot stat shared object: Stale file handle" in msg
                    or "No such file or directory"
                    and "/.triton/cache/" in msg
                ):
                    time.sleep(random.random() * 2 + 1)
                    continue
                raise e

    return wrapped


@auto_retry_triton_compilation
def append_to_paged_kv_cache(
    kv_cache,  # (num_pages, page_size, other dims...)
    page_table,  # (batch_size, num_pages_per_sample)
    this_kv,  # (batch_size, other dims...)
    old_seq_lens,  # (batch_size,)
):
    """
    for i in range(cache_seqlens.shape[0]):
        kv_cache[block_table[i][cache_seqlens[i] // 64]][cache_seqlens[i] % 64] = kv[i]
    """

    assert kv_cache.is_contiguous()
    assert page_table.is_contiguous()
    assert this_kv.is_contiguous()
    assert old_seq_lens.is_contiguous()

    page_size = kv_cache.shape[1]

    batch_size, num_pages_per_sample = page_table.shape
    assert this_kv.shape[0] == batch_size
    assert old_seq_lens.shape[0] == batch_size

    tot_len_of_other_dims = this_kv.numel() // batch_size
    assert (
        kv_cache.numel() // (kv_cache.shape[0] * kv_cache.shape[1])
        == tot_len_of_other_dims
    )

    block_size = 512  # GPU block size, not page size
    grid = (batch_size, triton.cdiv(tot_len_of_other_dims, block_size))
    append_to_paged_kv_cache_kernel[grid](
        kv_cache_ptr=kv_cache,
        page_table_ptr=page_table,
        this_kv_ptr=this_kv,
        old_seq_lens_ptr=old_seq_lens,
        PAGE_SIZE=page_size,
        BATCH_SIZE=batch_size,
        NUM_PAGES_PER_SAMPLE=num_pages_per_sample,
        TOT_LEN_OF_OTHER_DIMS=tot_len_of_other_dims,
        BLOCK_SIZE=block_size,
    )


def rotate_half(x):
    x1 = x[..., : x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2 :]
    return torch.cat((-x2, x1), dim=-1)


def rotate_pairwise(x):
    y = x.reshape(x.shape[:-1] + (x.shape[-1] // 2, 2))
    y = torch.cat((-y[..., 1:], y[..., :1]), dim=-1)
    return y.reshape(x.shape)


def reshape_rotary_for_broadcast(freqs_cis: torch.Tensor, x: torch.Tensor):
    assert freqs_cis.shape == (
        x.shape[0],
        x.shape[-1],
    ), f"{freqs_cis.shape} {x.shape}"
    ndim = x.ndim
    if ndim == 4:
        shape = [1, x.shape[1], 1, x.shape[-1]]
    elif ndim == 3:
        shape = [x.shape[0], 1, x.shape[-1]]
    elif ndim == 2:
        shape = [x.shape[0], x.shape[-1]]
    else:
        assert False
    return freqs_cis.view(*shape)


@auto_retry_triton_compilation
def apply_rotary_pos_emb_triton(q, k, cos, sin, rotary_type="hf-llama", block_size=128):
    if rotary_type == "hf-llama":
        # "hf-llama" has an [real, real, ..., real, imag, imag, ..., imag] layout.

        # Get tensor shapes
        q_batch_size, q_n_local_heads, q_head_dim = q.shape
        k_batch_size, k_n_local_heads, k_head_dim = k.shape

        # Prepare output tensor
        q_output = torch.empty_like(q)
        k_output = torch.empty_like(k)

        # Define grid size
        q_grid = (q_batch_size * q_n_local_heads, q_head_dim // block_size)
        k_grid = (k_batch_size * k_n_local_heads, k_head_dim // block_size)

        # Launch kernel (TODO: use only 1 kernel)
        assert q.is_contiguous()
        assert k.is_contiguous()
        assert cos.is_contiguous()
        assert sin.is_contiguous()
        assert q_output.is_contiguous()
        assert k_output.is_contiguous()
        rotary_embedding_kernel_hf_llama[q_grid](
            q,
            cos,
            sin,
            q_output,
            q_n_local_heads,
            q.stride(0),
            q.stride(1),
            cos.stride(0),
            sin.stride(0),
            q_output.stride(0),
            q_output.stride(1),
            BLOCK_SIZE=block_size,
        )
        rotary_embedding_kernel_hf_llama[k_grid](
            k,
            cos,
            sin,
            k_output,
            k_n_local_heads,
            k.stride(0),
            k.stride(1),
            cos.stride(0),
            sin.stride(0),
            k_output.stride(0),
            k_output.stride(1),
            BLOCK_SIZE=block_size,
        )

        return q_output, k_output

    elif rotary_type == "llama":
        # "llama" has an [real, imag, real, imag, ..., real, imag] layout.

        q_shape = q.shape
        k_shape = k.shape

        if q.dim() == 4:
            q = q.view(-1, q.shape[-2], q, shape[-1])
        elif q.dim() == 3:
            pass
        elif q.dim() == 2:
            q = q.view(-1, 1, q.shape[-1])
        else:
            assert False
        if k.dim() == 4:
            k = k.view(-1, k.shape[-2], k.shape[-1])
        elif k.dim() == 3:
            pass
        elif k.dim() == 2:
            k = k.view(-1, 1, k.shape[-1])
        else:
            assert False
        assert q.shape[-1] == k.shape[-1]
        assert q.shape[0] == k.shape[0]
        assert q.shape[-1] // 2 == cos.shape[-1]
        assert q.shape[-1] // 2 == sin.shape[-1]
        bs, head_num_q, rotary_dim = q.shape
        bs, head_num_k, rotary_dim = k.shape

        # Prepare output tensor
        out_q = torch.empty_like(q)
        out_k = torch.empty_like(k)

        # Launch kernel
        BLOCK_H = min(
            triton.cdiv(triton.next_power_of_2(bs), 128), max(head_num_q, head_num_k)
        )
        grid = lambda meta: (bs, triton.cdiv(max(head_num_q, head_num_k), BLOCK_H), 1)
        rotary_embedding_kernel_llama[grid](
            q,
            k,
            out_q,
            out_k,
            cos,
            sin,
            q.stride(0),
            q.stride(1),
            k.stride(0),
            k.stride(1),
            out_q.stride(0),
            out_q.stride(1),
            out_k.stride(0),
            out_k.stride(1),
            head_num_q,
            head_num_k,
            rotary_dim,
            BLOCK_H,
        )

        return out_q.view(q_shape), out_k.view(k_shape)

    else:
        raise ValueError(f"Unknown rotary type: {rotary_type}")


def apply_rotary_pos_emb_torch(q, k, cos, sin, rotary_type="hf-llama"):
    if rotary_type == "hf-llama":
        # "hf-llama" has an [real, real, ..., real, imag, imag, ..., imag] layout.
        cos = torch.cat([cos, cos], dim=-1)
        sin = torch.cat([sin, sin], dim=-1)
        cos_q = reshape_rotary_for_broadcast(cos, q)
        sin_q = reshape_rotary_for_broadcast(sin, q)
        cos_k = reshape_rotary_for_broadcast(cos, k)
        sin_k = reshape_rotary_for_broadcast(sin, k)
        q_embed = (q * cos_q) + (rotate_half(q) * sin_q)
        k_embed = (k * cos_k) + (rotate_half(k) * sin_k)
        return q_embed.to(q.dtype), k_embed.to(k.dtype)

    elif rotary_type == "llama":
        # "llama" has an [real, imag, real, imag, ..., real, imag] layout.
        cos = torch.stack([cos, cos], dim=-1).flatten(-2)
        sin = torch.stack([sin, sin], dim=-1).flatten(-2)
        cos_q = reshape_rotary_for_broadcast(cos, q)
        sin_q = reshape_rotary_for_broadcast(sin, q)
        cos_k = reshape_rotary_for_broadcast(cos, k)
        sin_k = reshape_rotary_for_broadcast(sin, k)
        q_embed = (q * cos_q) + (rotate_pairwise(q) * sin_q)
        k_embed = (k * cos_k) + (rotate_pairwise(k) * sin_k)
        return q_embed.to(q.dtype), k_embed.to(k.dtype)

    elif rotary_type == "glm4":
        # TODO: Now we transpose q and k, do the rotary, and transpose back.
        # Maybe we can transpose cos and sin just once instead of transposing q and k.
        q, q_pass = q[..., :64], q[..., 64:]
        k, k_pass = k[..., :64], k[..., 64:]
        q = (
            q.reshape(q.shape[0], q.shape[1], q.shape[2] // 2, 2)
            .permute(0, 1, 3, 2)
            .reshape(q.shape[0], q.shape[1], q.shape[2])
        )
        k = (
            k.reshape(k.shape[0], k.shape[1], k.shape[2] // 2, 2)
            .permute(0, 1, 3, 2)
            .reshape(k.shape[0], k.shape[1], k.shape[2])
        )
        cos = torch.stack([cos, cos], dim=-1).flatten(-2)
        sin = torch.stack([sin, sin], dim=-1).flatten(-2)
        cos = reshape_rotary_for_broadcast(cos, q)
        sin = reshape_rotary_for_broadcast(sin, q)
        q_embed = (q * cos) + (rotate_half(q) * sin)
        k_embed = (k * cos) + (rotate_half(k) * sin)
        q_embed = (
            q_embed.reshape(
                q_embed.shape[0], q_embed.shape[1], 2, q_embed.shape[2] // 2
            )
            .permute(0, 1, 3, 2)
            .reshape(q_embed.shape[0], q_embed.shape[1], q_embed.shape[2])
        )
        k_embed = (
            k_embed.reshape(
                k_embed.shape[0], k_embed.shape[1], 2, k_embed.shape[2] // 2
            )
            .permute(0, 1, 3, 2)
            .reshape(k_embed.shape[0], k_embed.shape[1], k_embed.shape[2])
        )
        return torch.cat([q_embed, q_pass], dim=-1), torch.cat(
            [k_embed, k_pass], dim=-1
        )

    else:
        raise ValueError(f"Unknown rotary type: {rotary_type}")


def apply_rotary_pos_emb(q, k, cos, sin, rotary_type="hf-llama"):
    if rotary_type == "hf-llama" or (
        rotary_type == "llama" and hasattr(triton.language, "interleave")
    ):
        # NOTE: some platform such as muxi now doesn't support triton.language.interleave, so we need check attr
        # NOTE: Performance of triton rotary kernel is untested for large batch sizes.
        # If it's slow on prefill, just switch to torch implementation on the else case.
        return apply_rotary_pos_emb_triton(q, k, cos, sin, rotary_type=rotary_type)
    else:
        return apply_rotary_pos_emb_torch(
            q,
            k,
            cos,
            sin,
            rotary_type=rotary_type,
        )


@auto_retry_triton_compilation
def act_quant_deepseek_v3(
    x: torch.Tensor, block_size: int = 128
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Quantizes the input tensor `x` using block-wise quantization.

    Args:
        x (torch.Tensor): The input tensor to be quantized. Must be contiguous and its last dimension size must be divisible by `block_size`.
        block_size (int, optional): The size of the blocks to be used for quantization. Default is 128.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: A tuple containing:
            - The quantized tensor with dtype `torch.float8_e4m3fn`.
            - A tensor of scaling factors with dtype `torch.float32`.
    """
    assert x.is_contiguous(), "Input tensor must be contiguous"
    assert (
        x.size(-1) % block_size == 0
    ), f"Last dimension size must be divisible by block_size (block_size={block_size})"
    y = torch.empty_like(x, dtype=torch.float8_e4m3fn)
    s = x.new_empty(*x.size()[:-1], x.size(-1) // block_size, dtype=torch.float32)
    grid = lambda meta: (triton.cdiv(x.numel(), meta["BLOCK_SIZE"]),)
    act_quant_deepseek_v3_kernel[grid](x, y, s, BLOCK_SIZE=block_size)
    return y, s


@auto_retry_triton_compilation
def weight_dequant_deepseek_v3(
    x: torch.Tensor, s: torch.Tensor, block_size: int = 128
) -> torch.Tensor:
    """
    Dequantizes the given weight tensor using the provided scale tensor.

    Args:
        x (torch.Tensor): The quantized weight tensor of shape (M, N).
        s (torch.Tensor): The scale tensor of shape (M / block_size, N / block_size).
        block_size (int, optional): The block size to use for dequantization. Defaults to 128.

    Returns:
        torch.Tensor: The dequantized weight tensor of the same shape as `x`.

    Raises:
        AssertionError: If `x` or `s` are not contiguous or if their dimensions are not 2.
    """
    assert x.is_contiguous() and s.is_contiguous(), "Input tensors must be contiguous"
    assert (
        s.dim() == x.dim()
    ), "Scale tensors must have the same number of dimensions with the weight tensor"
    if x.dim() == 2:
        M, N = x.size()
        B = 1
    elif x.dim() == 3:
        B, M, N = x.size()
    else:
        assert False, "Weight tensor must have 2 or 3 dimensions"
    y = torch.empty_like(x, dtype=torch.get_default_dtype())
    grid = lambda meta: (
        B,
        triton.cdiv(M, meta["BLOCK_SIZE"]),
        triton.cdiv(N, meta["BLOCK_SIZE"]),
    )
    weight_dequant_deepseek_v3_kernel[grid](x, s, y, M, N, BLOCK_SIZE=block_size)
    return y


@auto_retry_triton_compilation
def weight_dequant_soft_fp8_deepseek_v3(
    x: torch.Tensor, s: torch.Tensor, block_size: int = 128
) -> torch.Tensor:
    """
    Dequantizes the given weight tensor using the provided scale tensor.

    Args:
        x (torch.Tensor): The quantized weight tensor of shape (M, N).
        s (torch.Tensor): The scale tensor of shape (M / block_size, N / block_size).
        block_size (int, optional): The block size to use for dequantization. Defaults to 128.

    Returns:
        torch.Tensor: The dequantized weight tensor of the same shape as `x`.

    Raises:
        AssertionError: If `x` or `s` are not contiguous or if their dimensions are not 2.
    """
    assert x.is_contiguous() and s.is_contiguous(), "Input tensors must be contiguous"
    assert (
        s.dim() == x.dim()
    ), "Scale tensors must have the same number of dimensions with the weight tensor"
    if x.dim() == 2:
        M, N = x.size()
        B = 1
    elif x.dim() == 3:
        B, M, N = x.size()
    else:
        assert False, "Weight tensor must have 2 or 3 dimensions"

    x = x.view(dtype=torch.uint8)
    bit_reordered_x = torch.empty_like(x, dtype=torch.uint32)
    grid = lambda meta: (triton.cdiv(B * M * N, meta["BLOCK_SIZE"]),)
    weight_dequant_soft_fp8_deepseek_v3_kernel_step_1[grid](
        x, bit_reordered_x, B * M * N, BLOCK_SIZE=block_size
    )
    bit_reordered_x = bit_reordered_x.view(dtype=torch.float32)

    # Some of our platforms only has Triton with low versions, where these is no `tl.cast`
    # which is used for initializing a constant with a given type. Therefore, we need to
    # pass `fp8_to_fp32_scale` as a constant from outside.
    fp8_to_fp32_scale = struct.unpack(">f", bytes.fromhex("7b800000"))[0]
    y = torch.empty_like(x, dtype=torch.get_default_dtype())
    grid = lambda meta: (
        B,
        triton.cdiv(M, meta["BLOCK_SIZE"]),
        triton.cdiv(N, meta["BLOCK_SIZE"]),
    )
    weight_dequant_soft_fp8_deepseek_v3_kernel_step_2[grid](
        bit_reordered_x,
        s,
        y,
        M,
        N,
        BLOCK_SIZE=block_size,
        fp8_to_fp32_scale=fp8_to_fp32_scale,
    )
    return y


@auto_retry_triton_compilation
def fp8_gemm_deepseek_v3(
    a: torch.Tensor, a_s: torch.Tensor, b: torch.Tensor, b_s: torch.Tensor
):
    """
    Perform a matrix multiplication using FP8 precision.

    Args:
        a (torch.Tensor): The first input matrix, must be contiguous.
        a_s (torch.Tensor): The scaling factor for the first input matrix, must be contiguous.
        b (torch.Tensor): The second input matrix, must be contiguous.
        b_s (torch.Tensor): The scaling factor for the second input matrix, must be contiguous.

    Returns:
        torch.Tensor: The result of the matrix multiplication.
    """
    assert a.is_contiguous() and b.is_contiguous(), "Input tensors must be contiguous"
    assert (
        a_s.is_contiguous() and b_s.is_contiguous()
    ), "Scaling factor tensors must be contiguous"
    K = a.size(-1)
    M = a.numel() // K
    N = b.size(0)
    c = a.new_empty(*a.size()[:-1], N, dtype=torch.get_default_dtype())
    grid = lambda META: (
        triton.cdiv(M, META["BLOCK_SIZE_M"]),
        triton.cdiv(N, META["BLOCK_SIZE_N"]),
    )
    fp8_gemm_deepseek_v3_kernel[grid](
        a, b, c, a_s, b_s, M, N, K, group_n=128, group_k=128
    )
    return c


@auto_retry_triton_compilation
def soft_fp8_gemm_deepseek_v3(
    a: torch.Tensor,
    b: torch.Tensor,
    b_s: torch.Tensor,
):
    """
    Perform a matrix multiplication with FP8 dynamically casted to BF16.

    Args:
        a (torch.Tensor): The first input matrix, must be contiguous.
        b (torch.Tensor): The second input matrix, must be contiguous.
        b_s (torch.Tensor): The scaling factor for the second input matrix, must be contiguous.

    Returns:
        torch.Tensor: The result of the matrix multiplication.
    """
    assert a.is_contiguous() and b.is_contiguous(), "Input tensors must be contiguous"
    assert b_s.is_contiguous(), "Scaling factor tensor must be contiguous"
    K = a.size(-1)
    M = a.numel() // K
    N = b.size(0)
    c = a.new_empty(*a.size()[:-1], N, dtype=torch.get_default_dtype())

    # Some of our platforms only has Triton with low versions, where these is no `tl.cast`
    # which is used for initializing a constant with a given type. Therefore, we need to
    # pass `fp8_to_fp32_scale` as a constant from outside.
    fp8_to_fp32_scale = struct.unpack(">f", bytes.fromhex("7b800000"))[0]
    if torch.get_default_dtype() == torch.bfloat16:
        compute_dtype = tl.bfloat16
    elif torch.get_default_dtype() == torch.float16:
        compute_dtype = tl.float16
    elif torch.get_default_dtype() == torch.float32:
        compute_dtype = tl.float32
    else:
        raise ValueError(f"Unsupported compute_type: {torch.get_default_dtype()}")
    grid = lambda META: (
        triton.cdiv(M, META["BLOCK_SIZE_M"]) * triton.cdiv(N, META["BLOCK_SIZE_N"]),
    )
    soft_fp8_gemm_deepseek_v3_kernel[grid](
        a,
        b.view(dtype=torch.uint8),
        c,
        b_s,
        M,
        N,
        K,
        group_n=128,
        group_k=128,
        fp8_to_fp32_scale=fp8_to_fp32_scale,
        compute_dtype=compute_dtype,
    )
    return c



from logging import getLogger

import torch

from chitu.global_vars import get_slot_handle, get_timers, get_global_args
from chitu.static_tensor import StaticTensor

logger = getLogger(__name__)
_BLOCK_SIZE = 512  # _BLOCK_SIZE must be a multiple of 256 for FlashAttention
_MAX_SEQ_LEN = 2048


class PagedKVCacheManager:
    def __init__(
        self,
        begin_layer_id,
        end_layer_id,
        num_hot_req=16,
        block_size=_BLOCK_SIZE,
        max_seq_len=_MAX_SEQ_LEN,
        device="cuda",
        *,
        k_shape_per_sample=None,
        v_shape_per_sample=None,
        kv_shape_per_sample=None,
        n_local_kv_heads=None,
        head_dim=None,
    ):
        """
        You can optionally set `k_shae_per_sample` and `v_shape_per_sample`, or `n_local_kv_heads` and `head_dim`.
        """

        self.max_blocks_per_req = max_seq_len // block_size + 1
        self.num_blocks = self.max_blocks_per_req * num_hot_req

        self.begin_layer_id = begin_layer_id
        self.end_layer_id = end_layer_id
        self.num_layers = end_layer_id - begin_layer_id
        self.k_shape_per_sample = (
            k_shape_per_sample
            if k_shape_per_sample is not None
            else (n_local_kv_heads, head_dim)
        )
        self.v_shape_per_sample = (
            v_shape_per_sample
            if v_shape_per_sample is not None
            else (n_local_kv_heads, head_dim)
        )
        self.kv_shape_per_sample = kv_shape_per_sample
        # assert block_size % 256 == 0
        self.block_size = block_size
        self.max_seq_len = max_seq_len
        self.device = torch.device(device)

        self.seq_lens = {}
        self.timers = get_timers()
        self.block_table = {}  # (seq_id, block_idx)
        self.curr_seq_lens_gpu_excl_this_decode = StaticTensor(
            max_nelem=num_hot_req, dtype=torch.int32, device=self.device
        )
        self.curr_seq_lens_gpu_incl_this_decode = StaticTensor(
            max_nelem=num_hot_req, dtype=torch.int32, device=self.device
        )
        self.gpu_block_table = StaticTensor(
            max_nelem=self.num_blocks, dtype=torch.int32, device=self.device
        )
        # TODO: For better performance, use list instead of set for free_blocks
        self.free_blocks = set(range(self.num_blocks))
        if self.kv_shape_per_sample is not None:
            self.paged_kv_cache = torch.zeros(
                (self.num_layers, self.num_blocks, block_size)
                + self.kv_shape_per_sample,
                device=device,
            )
        else:
            self.paged_k_cache = torch.zeros(
                (self.num_layers, self.num_blocks, block_size)
                + self.k_shape_per_sample,
                device=device,
            )
            self.paged_v_cache = torch.zeros(
                (self.num_layers, self.num_blocks, block_size)
                + self.v_shape_per_sample,
                device=device,
            )

    def get_block_size(self):
        return self.block_size

    def get_num_blocks(self):
        return self.num_blocks

    # Init block table and kv cache with kv generated during prefill
    def finalize_cache_bylayer_prefill(self, xk, xv, req_ids, varlen, layer_id):
        self.timers("cache_finalize_cache_all_prefill").start()
        for idx, req_id in enumerate(req_ids):
            num_blocks_prepared = (
                varlen.cpu_lens[idx] + self.block_size - 1
            ) // self.block_size

            if layer_id == self.begin_layer_id:
                self.seq_lens[req_id] = varlen.cpu_lens[idx]
                block_ids = []
                for chunck_id in range(num_blocks_prepared):
                    block_idx = self.get_free_block()
                    block_ids.append(block_idx)
                self.block_table[req_id] = block_ids
            else:
                block_ids = self.block_table[req_id]

            start_pos = varlen.cpu_prefix_lens[idx]
            end_pos = varlen.cpu_prefix_lens[idx + 1]
            for chunck_id in range(num_blocks_prepared):
                block_idx = block_ids[chunck_id]
                if self.kv_shape_per_sample is not None:
                    if chunck_id != num_blocks_prepared - 1:
                        self.paged_kv_cache[layer_id - self.begin_layer_id][
                            block_idx
                        ] = xk[start_pos : (start_pos + self.block_size)].clone()
                        start_pos += self.block_size
                    else:
                        tmp_len = end_pos - start_pos
                        self.paged_kv_cache[layer_id - self.begin_layer_id][block_idx][
                            :tmp_len
                        ] = xk[start_pos:end_pos].clone()
                else:
                    if chunck_id != num_blocks_prepared - 1:
                        self.paged_k_cache[layer_id - self.begin_layer_id][
                            block_idx
                        ] = xk[start_pos : (start_pos + self.block_size)].clone()
                        self.paged_v_cache[layer_id - self.begin_layer_id][
                            block_idx
                        ] = xv[start_pos : (start_pos + self.block_size)].clone()
                        start_pos += self.block_size
                    else:
                        tmp_len = end_pos - start_pos
                        self.paged_k_cache[layer_id - self.begin_layer_id][block_idx][
                            :tmp_len
                        ] = xk[start_pos:end_pos].clone()
                        self.paged_v_cache[layer_id - self.begin_layer_id][block_idx][
                            :tmp_len
                        ] = xv[start_pos:end_pos].clone()
        self.timers("cache_finalize_cache_all_prefill").stop()

    def finalize_cache_all_prefill(self, req_ids, varlen):
        self.curr_varlens = None
        self.curr_req_ids = None

    def prepare_cache_decode(self, req_ids):
        seq_lens = []
        for req_id in req_ids:
            seq_len = self.seq_lens[req_id]
            seq_lens.append(seq_len)
        max_seq = max(seq_lens)
        self.curr_seq_lens = seq_lens
        self.curr_seq_lens_gpu_excl_this_decode.set(
            torch.tensor(seq_lens, dtype=torch.int32, device=self.device)
        )
        self.curr_seq_lens_gpu_incl_this_decode.set(
            self.curr_seq_lens_gpu_excl_this_decode.get() + 1
        )

    def get_free_block(self):
        # TODO: When run out of free blocks, use scheduling and preemption in paper instead of exception
        self.timers("get_free_block").start()
        if len(self.free_blocks) == 0:
            raise Exception("No more free blocks.")
        idx = list(self.free_blocks)[0]
        self.free_blocks.remove(idx)
        self.timers("get_free_block").stop()
        return idx

    def get_gpu_block_table(self):
        return self.gpu_block_table.get()

    def get_gpu_seq_lens_excl_this_decode(self):
        return self.curr_seq_lens_gpu_excl_this_decode.get()

    def get_gpu_seq_lens_incl_this_decode(self):
        return self.curr_seq_lens_gpu_incl_this_decode.get()

    def get_paged_kv_cache(self, layer_id):
        if self.kv_shape_per_sample is not None:
            return self.paged_kv_cache[layer_id - self.begin_layer_id]
        else:
            return (
                self.paged_k_cache[layer_id - self.begin_layer_id],
                self.paged_v_cache[layer_id - self.begin_layer_id],
            )

    def free_req_cache_blocks(self, req_id):
        self.timers("free_req_cache_blocks").start()
        for block in self.block_table[req_id]:
            self.free_blocks.add(block)
        del self.block_table[req_id]
        self.timers("free_req_cache_blocks").stop()

    # Prepare enough block table for next decoding. When decoding, flash attention will fill new kv into paged kv cache (inplace).
    def prepare_block_table_for_decode(self, req_ids):
        for req_id in req_ids:
            if self.seq_lens[req_id] % self.block_size == 0:
                self.block_table[req_id].append(self.get_free_block())

        if get_global_args().infer.use_cuda_graph:
            max_block_num = self.max_blocks_per_req
        else:
            max_block_num = max(len(self.block_table[req_id]) for req_id in req_ids)
        self.gpu_block_table.set(
            torch.zeros((len(req_ids), max_block_num), dtype=torch.int32, device="cuda")
        )
        for idx, req_id in enumerate(req_ids):
            block_ids = self.block_table[req_id]
            self.gpu_block_table.get()[idx, : len(block_ids)] = torch.tensor(
                block_ids, dtype=torch.int32
            )

    def finalize_cache_single_decode(self, req_ids):
        for req_id in req_ids:
            self.seq_lens[req_id] = self.seq_lens[req_id] + 1
        self.curr_varlens = None
        self.curr_req_ids = None

    def finalize_cache_all_decode(self, req_id):
        self.timers("finalize_cache_all_decode").start()
        assert req_id in self.seq_lens
        assert req_id in self.block_table
        del self.seq_lens[req_id]
        self.free_req_cache_blocks(req_id)
        self.curr_varlens = None
        self.curr_req_ids = None
        self.timers("finalize_cache_all_decode").stop()


class KVCacheManager:
    def __init__(
        self, begin_layer_id, end_layer_id, n_local_kv_heads, head_dim, device="cuda"
    ):
        self.cache = {}
        self.prepared_cache = []
        self.begin_layer_id = begin_layer_id
        self.end_layer_id = end_layer_id
        self.num_layers = end_layer_id - begin_layer_id
        self.n_local_kv_heads = n_local_kv_heads
        self.head_dim = head_dim
        self.tmp_storage = []
        self.seq_lens = {}
        self.timers = get_timers()
        self.device = torch.device(device)

    # Prefill:
    def finalize_cache_bylayer_prefill(
        self, cache_k, cache_v, req_ids, varlen, layer_id
    ):
        self.tmp_storage.append([cache_k, cache_v])

    # Prefill:
    # return for every req [layer, seq, n_local_kv_heads, head_dim] * 2 (for k and v)
    def finalize_cache_all_prefill(self, req_ids, varlen):
        self.timers("cache_finalize_cache_all_prefill").start()
        assert (
            len(self.tmp_storage) == self.num_layers
        ), f"{len(self.tmp_storage)} {self.num_layers}"
        assert len(varlen.cpu_lens) == len(req_ids)
        assert sum(varlen.cpu_lens) == self.tmp_storage[0][0].shape[0]
        for it, req_id in enumerate(req_ids):
            self.seq_lens[req_id] = varlen.cpu_lens[it]
            self.cache[req_id] = [None] * self.num_layers
        for local_layer_id in range(self.num_layers):
            start = 0
            for it, req_id in enumerate(req_ids):
                end = start + varlen.cpu_lens[it]
                self.cache[req_id][local_layer_id] = [
                    self.tmp_storage[local_layer_id][0][
                        start:end
                    ],  # [seq, n_local_kv_heads, head_dim]
                    self.tmp_storage[local_layer_id][1][
                        start:end
                    ],  # [seq, n_local_kv_heads, head_dim]
                ]
                start = end
        self.tmp_storage = []
        self.curr_varlens = None
        self.curr_req_ids = None
        self.timers("cache_finalize_cache_all_prefill").stop()

    # Decode:
    # return [layer, num_req, 2, max_seqlen + 1, n_local_kv_heads, head_dim]
    def prepare_cache_decode(self, req_ids):
        self.timers("cache_prepare").start()
        max_seq = 0
        seq_lens = []
        for req_id in req_ids:
            seq_len = self.seq_lens[req_id]
            seq_lens.append(seq_len)
        self.curr_seq_lens = seq_lens
        self.curr_seq_lens_gpu_excl_this_decode = torch.tensor(
            seq_lens, dtype=torch.int32, device=self.device
        )
        self.curr_seq_lens_gpu_incl_this_decode = (
            self.curr_seq_lens_gpu_excl_this_decode + 1
        )
        max_seq = max(seq_lens)
        n_local_kv_heads = self.cache[req_ids[0]][0][0].shape[-2]
        head_dim = self.cache[req_ids[0]][0][0].shape[-1]
        prepared_cache = torch.zeros(
            [
                self.num_layers,  # layers
                2,
                len(req_ids),  # batch_size
                max_seq + 1,  # seq_len
                n_local_kv_heads,  # n_local_kv_heads
                head_dim,  # head_dim
            ],
            device=self.device,
        )
        # hkz-comment: Very similar to matrix transpose;
        for local_layer_id in range(self.num_layers):
            for it, req_id in enumerate(req_ids):
                prepared_cache[local_layer_id][0][it][: seq_lens[it]] = self.cache[
                    req_id
                ][local_layer_id][0]
                prepared_cache[local_layer_id][1][it][: seq_lens[it]] = self.cache[
                    req_id
                ][local_layer_id][1]
        self.prepared_cache = prepared_cache
        self.timers("cache_prepare").stop()

    # Decode:
    # return [2, num_req, max_seqlen + 1, n_local_kv_heads, head_dim]
    def get_cache_decode(self, layer_id):
        return self.prepared_cache[layer_id - self.begin_layer_id]

    def get_gpu_block_table(self):
        return None

    def get_block_size(self):
        return 0

    # Decode:
    # return [# of current req_ids]
    def get_gpu_seq_lens_excl_this_decode(self):
        return self.curr_seq_lens_gpu_excl_this_decode

    # Decode:
    # return [# of current req_ids]
    def get_gpu_seq_lens_incl_this_decode(self):
        return self.curr_seq_lens_gpu_incl_this_decode

    # Decode:
    # return for every req [layer, seq + 1, n_local_kv_heads, head_dim] * 2 (for k and v)
    def finalize_cache_single_decode(self, req_ids):
        self.timers("cache_finalize_cache_single_decode").start()
        assert len(self.prepared_cache) > 0
        for it, req_id in enumerate(req_ids):
            self.cache[req_id] = [None] * self.num_layers
            self.seq_lens[req_id] += 1
        for local_layer_id in range(self.num_layers):
            for it, req_id in enumerate(req_ids):
                self.cache[req_id][local_layer_id] = [
                    self.prepared_cache[local_layer_id][0][it][
                        : self.curr_seq_lens[it] + 1
                    ],  # [seq + 1, n_local_kv_heads, head_dim]
                    self.prepared_cache[local_layer_id][1][it][
                        : self.curr_seq_lens[it] + 1
                    ],  # [seq + 1, n_local_kv_heads, head_dim]
                ]
        self.prepared_cache = []
        self.timers("cache_finalize_cache_single_decode").stop()
        self.curr_varlens = None
        self.curr_req_ids = None

    # Decode:
    def finalize_cache_all_decode(self, req_id):
        self.curr_varlens = None
        self.curr_req_ids = None
        pass


class KVCache:
    def __init__(self):
        self.cache_k = None
        self.cache_v = None
        self.inited = False

    def check_shape(self, cache):
        # (bsz * seqlen, self.n_local_heads, self.head_dim)
        assert len(cache.shape) == 3, cache.shape
        assert cache.device != torch.device("cpu")

    def check_shapes(self, cache_k, cache_v):
        self.check_shape(cache_k)
        self.check_shape(cache_v)
        assert cache_k.shape == cache_v.shape

    def init(self, cache_k, cache_v):
        assert not self.inited
        # self.check_shapes(cache_k, cache_v)

        self.cache_k = cache_k
        self.cache_v = cache_v
        self.inited = True

    def extend(self, new_cache_k, new_cache_v):
        assert self.inited
        self.cache_k = torch.cat([self.cache_k, new_cache_k], dim=0)
        self.cache_v = torch.cat([self.cache_v, new_cache_v], dim=0)


class KVCacheManagerSkewAware:
    def __init__(
        self,
        begin_layer_id,
        end_layer_id,
        num_hot_req,
        max_seq_len=2048,
        device="cuda",
        *,
        k_shape_per_sample=None,
        v_shape_per_sample=None,
        n_local_kv_heads=None,
        head_dim=None,
    ):
        """
        You can optionally set `k_shae_per_sample` and `v_shape_per_sample`, or `n_local_kv_heads` and `head_dim`.
        """

        self.begin_layer_id = begin_layer_id
        self.end_layer_id = end_layer_id
        self.num_layers = end_layer_id - begin_layer_id
        self.k_shape_per_sample = (
            k_shape_per_sample
            if k_shape_per_sample is not None
            else (n_local_kv_heads, head_dim)
        )
        self.v_shape_per_sample = (
            v_shape_per_sample
            if v_shape_per_sample is not None
            else (n_local_kv_heads, head_dim)
        )
        self.num_hot_req = num_hot_req
        self.slot_availability = [True] * num_hot_req
        self.hot_reqs = [-1] * num_hot_req
        self.req2slot = {}
        self.seq_lens = {}
        self.max_seq_len = max_seq_len
        self.tmp_storage = []
        self.device = torch.device(device)
        self.k_buffer = torch.zeros(
            (
                self.num_layers,
                self.num_hot_req,
                self.max_seq_len,
            )
            + self.k_shape_per_sample,
            device=self.device,
        )
        self.v_buffer = torch.zeros(
            (
                self.num_layers,
                self.num_hot_req,
                self.max_seq_len,
            )
            + self.v_shape_per_sample,
            device=self.device,
        )
        self.timers = get_timers()
        self.prepared_reqs = []
        self.rounded_max_seq = -1
        self.slot_handle = get_slot_handle()

    # Prefill:
    # return for every req [layer, seq, n_local_kv_heads, head_dim] * 2 (for k and v)
    def finalize_cache_bylayer_prefill(
        self, cache_k, cache_v, req_ids, varlen, layer_id
    ):
        self.timers("cache_finalize_cache_all_prefill").start()
        if self.slot_handle:
            start_idx, _ = self.slot_handle.get_current_slot_start_end_idx()
        else:
            start_idx = 0
        if layer_id == self.begin_layer_id:
            for it, req_id in enumerate(req_ids):
                self.seq_lens[req_id] = varlen.cpu_lens[it]
                for i in range(start_idx, self.num_hot_req):
                    if self.slot_availability[i]:
                        self.req2slot[req_id] = i
                        self.slot_availability[i] = False
                        self.hot_reqs[i] = req_id
                        break
                assert (
                    req_id in self.req2slot
                ), f"Cannot allocate slot: {req_id} {self.req2slot}"

        start = 0
        for it, req_id in enumerate(req_ids):
            end = start + varlen.cpu_lens[it]
            self.k_buffer[layer_id - self.begin_layer_id][self.req2slot[req_id]][
                : varlen.cpu_lens[it]
            ] = cache_k[start:end]
            self.v_buffer[layer_id - self.begin_layer_id][self.req2slot[req_id]][
                : varlen.cpu_lens[it]
            ] = cache_v[start:end]
            start = end
        self.timers("cache_finalize_cache_all_prefill").stop()

    # Prefill:
    def finalize_cache_all_prefill(self, req_ids, varlen):
        pass

    # Decode:
    def prepare_cache_decode(self, req_ids):
        self.timers("cache_prepare").start()
        start_pos = self.hot_reqs.index(req_ids[0])
        assert start_pos + len(req_ids) <= self.num_hot_req
        # assert (
        #     self.hot_reqs[start_pos : start_pos + len(req_ids)] == req_ids
        # ), f"{self.hot_reqs} {req_ids}"

        seq_lens = []
        for req_id in req_ids:
            seq_len = self.seq_lens[req_id]
            seq_lens.append(seq_len)
        max_seq = max(seq_lens)
        self.curr_seq_lens = seq_lens
        self.curr_seq_lens_gpu_excl_this_decode = torch.tensor(
            seq_lens, dtype=torch.int32, device=self.device
        )
        self.curr_seq_lens_gpu_incl_this_decode = (
            self.curr_seq_lens_gpu_excl_this_decode + 1
        )

        limit = 16
        rounded_max_seq = (max_seq + 1 + limit - 1) // limit * limit
        if self.rounded_max_seq >= rounded_max_seq and self.prepared_reqs == req_ids:
            # prepared cache is long enough
            self.timers("cache_prepare").stop()
            return

        self.rounded_max_seq = rounded_max_seq
        self.prepared_reqs = req_ids

        k_prepared_cache_shape = list(self.k_buffer.shape)
        k_prepared_cache_stride = list(self.k_buffer.stride())
        k_prepared_cache_stride[0] = (
            k_prepared_cache_shape[1] * k_prepared_cache_stride[1]
        )
        k_prepared_cache_shape[1] = len(req_ids)
        k_prepared_cache_shape[2] = rounded_max_seq
        k_prepared_cache_offset = start_pos * k_prepared_cache_stride[1]
        self.k_prepared_cache = torch.as_strided(
            self.k_buffer,
            k_prepared_cache_shape,
            k_prepared_cache_stride,
            k_prepared_cache_offset,
        )

        v_prepared_cache_shape = list(self.v_buffer.shape)
        v_prepared_cache_stride = list(self.v_buffer.stride())
        v_prepared_cache_stride[0] = (
            v_prepared_cache_shape[1] * v_prepared_cache_stride[1]
        )
        v_prepared_cache_shape[1] = len(req_ids)
        v_prepared_cache_shape[2] = rounded_max_seq
        v_prepared_cache_offset = start_pos * v_prepared_cache_stride[1]
        self.v_prepared_cache = torch.as_strided(
            self.v_buffer,
            v_prepared_cache_shape,
            v_prepared_cache_stride,
            v_prepared_cache_offset,
        )

        self.timers("cache_prepare").stop()

    # Decode:
    # return [2, num_req, max_seqlen + 1, n_local_kv_heads, head_dim]
    def get_cache_decode(self, layer_id):
        return (
            self.k_prepared_cache[layer_id - self.begin_layer_id],
            self.v_prepared_cache[layer_id - self.begin_layer_id],
        )

    def get_gpu_block_table(self):
        return None

    def get_block_size(self):
        return 0

    # Decode:
    # return [# of current req_ids]
    def get_gpu_seq_lens_excl_this_decode(self):
        return self.curr_seq_lens_gpu_excl_this_decode

    # Decode:
    # return [# of current req_ids]
    def get_gpu_seq_lens_incl_this_decode(self):
        return self.curr_seq_lens_gpu_incl_this_decode

    # Decode:
    def finalize_cache_single_decode(self, req_ids):
        for item in req_ids:
            self.seq_lens[item] += 1
        self.curr_varlens = None
        self.curr_req_ids = None

    # Decode:
    def finalize_cache_all_decode(self, req_id):
        slot_id = self.hot_reqs.index(req_id)
        if slot_id == -1:  # not in the hot slot
            return

        if self.slot_handle:
            # get end_idx in req_id slot
            end_idx = 0
            slot_end_idx = self.slot_handle.slot_end_idx
            for idx in slot_end_idx:
                if slot_id < idx:
                    end_idx = idx
                    break
            assert end_idx > slot_id, f"get the wrong id in skewkvcache"
            slot_last_id = None
            for idx in range(end_idx - 1, slot_id, -1):
                if not self.slot_availability[idx]:
                    slot_last_id = idx
                    break
        else:
            slot_last_id = next(
                (
                    i
                    for i in range(slot_id + 1, self.num_hot_req)
                    if (
                        not self.slot_availability[i]
                        and (i + 1 >= self.num_hot_req or self.slot_availability[i + 1])
                    )
                ),
                None,
            )

        if slot_last_id is not None:
            self.k_buffer[:, slot_id] = self.k_buffer[:, slot_last_id]
            self.v_buffer[:, slot_id] = self.v_buffer[:, slot_last_id]
            req_key = next(
                (k for k, v in self.req2slot.items() if v == slot_last_id), None
            )
            if req_key is not None:
                self.req2slot[req_key] = slot_id
                self.hot_reqs[slot_id] = req_key
            self.hot_reqs[slot_last_id] = -1
            self.slot_availability[slot_last_id] = True
            if req_id in self.req2slot:
                self.req2slot.pop(req_id)
            self.k_buffer[:, slot_last_id].zero_()
            self.v_buffer[:, slot_last_id].zero_()
        else:
            self.hot_reqs[slot_id] = -1
            self.slot_availability[slot_id] = True
            self.req2slot.pop(req_id)
            self.k_buffer[:, slot_id].zero_()
            self.v_buffer[:, slot_id].zero_()


class KVCacheManagerNop:
    def __init__(
        self,
        begin_layer_id,
        end_layer_id,
        n_local_kv_heads,
        head_dim,
        num_hot_req=16,
        max_seq_len=2048,
        device="cuda",
    ):
        self.begin_layer_id = begin_layer_id
        self.end_layer_id = end_layer_id
        self.num_layers = end_layer_id - begin_layer_id
        self.n_local_kv_heads = n_local_kv_heads
        self.head_dim = head_dim
        self.num_hot_req = num_hot_req
        self.slot_availability = [True] * num_hot_req
        self.hot_reqs = [-1] * num_hot_req
        self.req2slot = {}
        self.max_seq_len = max_seq_len
        self.tmp_storage = []
        self.device = torch.device(device)
        self.buffer = torch.zeros(
            [
                self.num_layers,
                2,
                self.num_hot_req,
                self.max_seq_len,
                self.n_local_kv_heads,
                self.head_dim,
            ],
            device=self.device,
        )
        pass

    # Prefill:
    # return for every req [layer, seq, n_local_kv_heads, head_dim] * 2 (for k and v)
    def finalize_cache_bylayer_prefill(self, cache_k, cache_v, req_ids, varlen):
        pass

    # Prefill:
    def finalize_cache_all_prefill(self, req_ids, varlen):
        pass

    # Decode:
    def prepare_cache_decode(self, req_ids):
        pass

    def get_gpu_block_table(self):
        return None

    def get_block_size(self):
        return 0

    # Decode:
    # return [# of current req_ids]
    def get_gpu_seq_lens_excl_this_decode(self):
        pass

    # Decode:
    # return [# of current req_ids]
    def get_gpu_seq_lens_incl_this_decode(self):
        pass

    # Decode:
    # return [2, num_req, max_seqlen + 1, n_local_kv_heads, head_dim]
    def get_cache_decode(self, layer_id):
        pass

    # Decode:
    def finalize_cache_single_decode(self, req_ids):
        self.curr_varlens = None
        self.curr_req_ids = None

    # Decode:
    def finalize_cache_all_decode(self, req_id):
        pass



__all__ = [
    "init_tp",
    "get_tp_group",
    "get_tp_size",
    "get_tp_rank",
    "ColumnParallelLinear",
    "RowParallelLinear",
    "VocabParallelEmbedding",
]

import torch

tp_comm_group = None


def generate_tp_rank_list(tp_size: int, pp_size: int):
    return torch.arange(tp_size * pp_size).reshape(pp_size, tp_size).tolist()


def init_tp(tp_size: int, pp_size: int):
    global tp_comm_group
    rank_list = generate_tp_rank_list(tp_size, pp_size)
    global_rank = torch.distributed.get_rank()
    for ranks in rank_list:
        group = torch.distributed.new_group(ranks)
        if global_rank in ranks:
            tp_comm_group = group


def get_tp_group():
    return tp_comm_group


def get_tp_size():
    return tp_comm_group.size() if tp_comm_group is not None else 1


def get_tp_rank():
    return torch.distributed.get_rank(group=get_tp_group())


class ColumnParallelLinear(torch.nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        has_bias: bool = True,
        gather_output: bool = True,
        dtype=None,
        bias_dtype=None,
        linear_op=torch.nn.functional.linear,
    ):
        """
        Ouput-dimension-parallelized linaer layer

        Args:
            in_features: size of each input sample
            out_features: size of each output sample
            has_bias: If set to True, the layer will have a bias.
            gather_output: If set to True, an all-gather operation is performed on the output tensor.
            dtype: The desired data type of the parameters.
            bias_dtype: The desired data type of the bias. Defaults to `dtype`.
            linear_op: The linear operation to use. Defaults to `torch.nn.functional.linear`.
        """

        super().__init__()

        self.tp_group = get_tp_group()
        self.tp_size = get_tp_size()

        # These attributes are unused, but keep them compatible with nn.Linear
        self.in_features = in_features
        self.out_features = out_features

        assert (
            out_features % self.tp_size == 0
        ), "out_features must be divisible by tp_size"

        self.gather_output = gather_output
        self.linear_op = linear_op

        self.weight = torch.nn.Parameter(
            torch.empty(out_features // self.tp_size, in_features, dtype=dtype)
        )
        if has_bias:
            self.bias = torch.nn.Parameter(
                torch.empty(out_features // self.tp_size, dtype=bias_dtype or dtype)
            )
        else:
            self.bias = None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        y = self.linear_op(x, self.weight, self.bias)
        if self.gather_output and self.tp_size > 1:
            y_transposed = y.permute(-1, *range(y.dim() - 1)).contiguous()
            shape = list(y_transposed.shape)
            shape[0] *= self.tp_size
            y_gathered = y.new_empty(shape)
            torch.distributed.all_gather_into_tensor(
                y_gathered, y_transposed, group=self.tp_group
            )
            y = y_gathered.permute(*range(1, y.dim()), 0)
        return y


class RowParallelLinear(torch.nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        has_bias: bool = True,
        input_is_parallel: bool = False,
        dtype=None,
        bias_dtype=None,
        linear_op=torch.nn.functional.linear,
    ):
        """
        Input-dimension-parallelized linear layer

        Args:
            in_features: size of each input sample
            out_features: size of each output sample
            has_bias: If set to True, the layer will have a bias.
            input_is_parallel: If set to True, the input tensor is already parallelized.
            dtype: The desired data type of the parameters.
            bias_dtype: The desired data type of the bias. Defaults to `dtype`.
            linear_op: The linear operation to use. Defaults to `torch.nn.functional.linear`.
        """

        super().__init__()

        self.tp_group = get_tp_group()
        self.tp_size = get_tp_size()
        self.rank = get_tp_rank()

        # These attributes are unused, but keep them compatible with nn.Linear
        self.in_features = in_features
        self.out_features = out_features

        assert (
            in_features % self.tp_size == 0
        ), "in_features must be divisible by tp_size"

        self.input_is_parallel = input_is_parallel
        self.linear_op = linear_op

        self.weight = torch.nn.Parameter(
            torch.empty(out_features, in_features // self.tp_size, dtype=dtype)
        )
        if has_bias:
            self.bias = torch.nn.Parameter(
                torch.empty(out_features, dtype=bias_dtype or dtype)
            )
        else:
            self.bias = None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not self.input_is_parallel and self.tp_size > 1:
            shape = list(x.shape)
            this_rank_dim = shape[-1] // self.tp_size
            shape[-1] = self.tp_size
            shape.append(this_rank_dim)
            x = x.view(shape).select(-2, self.rank)
        if self.tp_size > 1:
            y = self.linear_op(x, self.weight, self.bias if self.rank == 0 else None)
            torch.distributed.all_reduce(y, group=self.tp_group)
        else:
            y = self.linear_op(x, self.weight, self.bias)
        return y


class VocabParallelEmbedding(torch.nn.Module):
    def __init__(self, num_embeddings: int, embedding_dim: int, dtype=None):
        """
        Parallelized embedding layer

        Args:
            num_embeddings: size of the dictionary of embeddings
            embedding_dim: the size of each embedding vector
            dtype: The desired data type of the parameters.
        """

        super().__init__()

        self.tp_group = get_tp_group()
        self.tp_size = get_tp_size()
        self.rank = get_tp_rank()

        assert (
            num_embeddings % self.tp_size == 0
        ), "num_embeddings must be divisible by tp_size"
        self.vocab_start_idx = self.rank * (num_embeddings // self.tp_size)
        self.vocab_end_idx = self.vocab_start_idx + (num_embeddings // self.tp_size)

        self.weight = torch.nn.Parameter(
            torch.empty(num_embeddings // self.tp_size, embedding_dim, dtype=dtype)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.tp_size > 1:
            mask = (x < self.vocab_start_idx) | (x >= self.vocab_end_idx)
            x = x - self.vocab_start_idx
            x[mask] = 0
        y = torch.nn.functional.embedding(x, self.weight)
        if self.tp_size > 1:
            y[mask] = 0
            torch.distributed.all_reduce(y, group=self.tp_group)
        return y



"""
This file has adaption of open-source code from the following sources:
- The Triton implementation of fused MoE kernel is originally from vLLM (https://github.com/vllm-project/vllm/blob/ed6e9075d31e32c8548b480a47d1ffb77da1f54c/vllm/model_executor/layers/fused_moe/fused_moe.py), licensed under Apache 2.0.
- The CUDA implementation to align block for MoE is originally from SGLang
  (https://github.com/sgl-project/sglang/commit/ba5112ff691d791a9e38c6c71f59324a5fcb49d0),
  licensed under Apache 2.0.
"""

import functools
import json
import os
import struct
from typing import Any, Callable, Dict, List, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

import triton
import triton.language as tl
import chitu_backend
from chitu.device_type import is_muxi, is_nvidia, get_device_name


class SiluAndMul(nn.Module):
    """An activation function for SwiGLU.

    The function computes x -> silu(x[:d]) * x[d:] where d = x.shape[-1] // 2.

    Shapes:
        x: (num_tokens, 2 * d) or (batch_size, seq_len, 2 * d)
        return: (num_tokens, d) or (batch_size, seq_len, d)
    """

    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        d = x.shape[-1] // 2
        return F.silu(x[..., :d]) * x[..., d:]


@triton.jit
def write_zeros_to_output(
    c_ptr,
    stride_cm,
    stride_cn,
    pid_n,
    N,
    offs_token,
    token_mask,
    BLOCK_SIZE_M,
    BLOCK_SIZE_N,
    compute_type,
):
    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=compute_type)
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    c_ptrs = c_ptr + stride_cm * offs_token[:, None] + stride_cn * offs_cn[None, :]
    c_mask = token_mask[:, None] & (offs_cn[None, :] < N)
    tl.store(c_ptrs, accumulator, mask=c_mask)


@triton.jit
def fused_moe_kernel(
    # Pointers to matrices
    a_ptr,
    b_ptr,
    c_ptr,
    a_scale_ptr,
    b_scale_ptr,
    topk_weights_ptr,
    sorted_token_ids_ptr,
    expert_ids_ptr,
    num_tokens_post_padded_ptr,
    # Matrix dimensions
    N,
    K,
    EM,
    num_valid_tokens,
    # The stride variables represent how much to increase the ptr by when
    # moving by 1 element in a particular dimension. E.g. `stride_am` is
    # how much to increase `a_ptr` by to get the element one row down
    # (A has M rows).
    stride_am,
    stride_ak,
    stride_be,
    stride_bk,
    stride_bn,
    stride_cm,
    stride_cn,
    stride_asm,
    stride_ask,
    stride_bse,
    stride_bsk,
    stride_bsn,
    # Block size for block-wise quantization
    group_n: tl.constexpr,
    group_k: tl.constexpr,
    # Meta-parameters
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    BLOCK_SIZE_K: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
    MUL_ROUTED_WEIGHT: tl.constexpr,
    top_k: tl.constexpr,
    compute_type: tl.constexpr,
    use_fp8_w8a8: tl.constexpr,
    use_int8_w8a16: tl.constexpr,
    soft_fp8: tl.constexpr,
    fp8_to_fp32_scale: tl.constexpr,
):
    """
    Implements the fused computation for a Mixture of Experts (MOE) using
    token and expert matrices.

    Key Parameters:
    - A: The input tensor representing tokens with shape (*, K), where '*' can
        be any shape representing batches and K is the feature dimension of
        each token.
    - B: The stacked MOE weight tensor with shape (E, N, K), where E is
        the number of experts, K is the input feature dimension, and N is
        the output feature dimension.
    - C: The output cache tensor with shape (M, topk, N), where M is the
        total number of tokens post padding, topk is the number of times
        each token is repeated, and N is the output feature dimension.
    - sorted_token_ids: A tensor containing the sorted indices of tokens,
        repeated topk times and arranged by the expert index they are
        assigned to.
    - expert_ids: A tensor containing the indices of the expert for each
        block. It determines which expert matrix from B should be used for
        each block in A.
    This kernel performs the multiplication of a token by its corresponding
    expert matrix as determined by `expert_ids`. The sorting of
    `sorted_token_ids` by expert index and padding ensures divisibility by
    BLOCK_SIZE_M, which is necessary to maintain consistency in block matrix
    multiplication across different blocks processed by the same expert.
    """
    # -----------------------------------------------------------
    # Map program ids `pid` to the block of C it should compute.
    # This is done in a grouped ordering to promote L2 data reuse.
    pid = tl.program_id(axis=0)
    num_pid_m = tl.cdiv(EM, BLOCK_SIZE_M)
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)
    num_pid_in_group = GROUP_SIZE_M * num_pid_n
    group_id = pid // num_pid_in_group
    first_pid_m = group_id * GROUP_SIZE_M
    group_size_m = min(num_pid_m - first_pid_m, GROUP_SIZE_M)
    pid_m = first_pid_m + ((pid % num_pid_in_group) % group_size_m)
    pid_n = (pid % num_pid_in_group) // group_size_m

    # ----------------------------------------------------------
    # Create pointers for the first blocks of A and B.
    # We will advance this pointer as we move in the K direction
    # and accumulate
    # `a_ptrs` is a block of [BLOCK_SIZE_M, BLOCK_SIZE_K] pointers
    # `b_ptrs` is a block of [BLOCK_SIZE_K, BLOCK_SIZE_N] pointers
    num_tokens_post_padded = tl.load(num_tokens_post_padded_ptr)
    if pid_m * BLOCK_SIZE_M >= num_tokens_post_padded:
        return
    offs_token_id = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M).to(tl.int64)
    offs_token = tl.load(sorted_token_ids_ptr + offs_token_id)
    token_mask = offs_token < num_valid_tokens

    off_experts = tl.load(expert_ids_ptr + pid_m).to(tl.int64)
    if off_experts == -1:
        # -----------------------------------------------------------
        # Write back zeros to the output when the expert is not
        # in the current expert parallel rank.
        write_zeros_to_output(
            c_ptr,
            stride_cm,
            stride_cn,
            pid_n,
            N,
            offs_token,
            token_mask,
            BLOCK_SIZE_M,
            BLOCK_SIZE_N,
            compute_type,
        )
        return

    offs_bn = (pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N).to(tl.int64)) % N
    offs_k = tl.arange(0, BLOCK_SIZE_K)
    a_ptrs = a_ptr + (
        offs_token[:, None] // top_k * stride_am + offs_k[None, :] * stride_ak
    )

    b_ptrs = (
        b_ptr
        + off_experts * stride_be
        + (offs_k[:, None] * stride_bk + offs_bn[None, :] * stride_bn)
    )
    if use_int8_w8a16:
        b_scale_ptrs = (
            b_scale_ptr + off_experts * stride_bse + offs_bn[None, :] * stride_bsn
        )
        b_scale = tl.load(b_scale_ptrs)

    if use_fp8_w8a8:
        if group_k > 0 and group_n > 0:
            if not soft_fp8:
                a_scale_ptrs = a_scale_ptr + (offs_token // top_k) * stride_asm
            offs_bsn = offs_bn // group_n
            b_scale_ptrs = (
                b_scale_ptr + off_experts * stride_bse + offs_bsn * stride_bsn
            )
        else:
            a_scale = tl.load(a_scale_ptr)
            b_scale = tl.load(b_scale_ptr + off_experts)

    # -----------------------------------------------------------
    # Iterate to compute a block of the C matrix.
    # We accumulate into a `[BLOCK_SIZE_M, BLOCK_SIZE_N]` block
    # of fp32 values for higher accuracy.
    # `accumulator` will be converted back to fp16 after the loop.
    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K)):
        # Load the next block of A and B, generate a mask by checking the
        # K dimension.
        a = tl.load(
            a_ptrs,
            mask=token_mask[:, None] & (offs_k[None, :] < K - k * BLOCK_SIZE_K),
            other=0.0,
        )
        b = tl.load(b_ptrs, mask=offs_k[:, None] < K - k * BLOCK_SIZE_K, other=0.0)
        # We accumulate along the K dimension.
        if use_int8_w8a16:
            accumulator = tl.dot(a, b.to(compute_type), acc=accumulator)
        elif use_fp8_w8a8:
            if group_k > 0 and group_n > 0:
                k_start = k * BLOCK_SIZE_K
                offs_ks = k_start // group_k
                b_scale = tl.load(b_scale_ptrs + offs_ks * stride_bsk)

                if soft_fp8:
                    b_uint32 = b.to(tl.uint8, bitcast=True).to(tl.uint32)
                    b_unscaled_fp32 = (
                        ((b_uint32 & 0x80) << 24) | ((b_uint32 & 0x7F) << 20)
                    ).to(tl.float32, bitcast=True)
                    b_new_scale = b_scale * fp8_to_fp32_scale
                    b_scaled_fp32 = b_unscaled_fp32 * b_new_scale
                    b_scaled_fp32 = b_scaled_fp32.to(dtype=compute_type)
                    accumulator += tl.dot(a, b_scaled_fp32)
                else:
                    a_scale = tl.load(
                        a_scale_ptrs + offs_ks * stride_ask, mask=token_mask, other=0.0
                    )
                    accumulator += tl.dot(a, b) * a_scale[:, None] * b_scale[None, :]
            else:
                accumulator = tl.dot(a, b, acc=accumulator)
        else:
            accumulator += tl.dot(a, b)
        # Advance the ptrs to the next K block.
        a_ptrs += BLOCK_SIZE_K * stride_ak
        b_ptrs += BLOCK_SIZE_K * stride_bk

    if MUL_ROUTED_WEIGHT:
        moe_weight = tl.load(topk_weights_ptr + offs_token, mask=token_mask, other=0)
        accumulator = accumulator * moe_weight[:, None]
    if use_int8_w8a16:
        accumulator = (accumulator * b_scale).to(compute_type)
    elif use_fp8_w8a8:
        if group_k > 0 and group_n > 0:
            accumulator = accumulator.to(compute_type)
        else:
            accumulator = (accumulator * a_scale * b_scale).to(compute_type)
    else:
        accumulator = accumulator.to(compute_type)
    # -----------------------------------------------------------
    # Write back the block of the output
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    c_ptrs = c_ptr + stride_cm * offs_token[:, None] + stride_cn * offs_cn[None, :]
    c_mask = token_mask[:, None] & (offs_cn[None, :] < N)
    tl.store(c_ptrs, accumulator, mask=c_mask)


def ceil_div(a, b):
    return (a + b - 1) // b


@triton.jit
def moe_align_block_size_stage1(
    topk_ids_ptr,
    tokens_cnts_ptr,
    num_experts: tl.constexpr,
    numel: tl.constexpr,
    tokens_per_thread: tl.constexpr,
):
    pid = tl.program_id(0)

    start_idx = pid * tokens_per_thread

    off_c = (pid + 1) * num_experts

    for i in range(tokens_per_thread):
        if start_idx + i < numel:
            idx = tl.load(topk_ids_ptr + start_idx + i)
            token_cnt = tl.load(tokens_cnts_ptr + off_c + idx)
            tl.store(tokens_cnts_ptr + off_c + idx, token_cnt + 1)


@triton.jit
def moe_align_block_size_stage2(
    tokens_cnts_ptr,
    num_experts: tl.constexpr,
):
    pid = tl.program_id(0)

    last_cnt = 0
    for i in range(1, num_experts + 1):
        token_cnt = tl.load(tokens_cnts_ptr + i * num_experts + pid)
        last_cnt = last_cnt + token_cnt
        tl.store(tokens_cnts_ptr + i * num_experts + pid, last_cnt)


@triton.jit
def moe_align_block_size_stage3(
    total_tokens_post_pad_ptr,
    tokens_cnts_ptr,
    cumsum_ptr,
    num_experts: tl.constexpr,
    block_size: tl.constexpr,
):
    last_cumsum = 0
    off_cnt = num_experts * num_experts
    for i in range(1, num_experts + 1):
        token_cnt = tl.load(tokens_cnts_ptr + off_cnt + i - 1)
        last_cumsum = last_cumsum + tl.cdiv(token_cnt, block_size) * block_size
        tl.store(cumsum_ptr + i, last_cumsum)
    tl.store(total_tokens_post_pad_ptr, last_cumsum)


@triton.jit
def moe_align_block_size_stage4(
    topk_ids_ptr,
    sorted_token_ids_ptr,
    expert_ids_ptr,
    tokens_cnts_ptr,
    cumsum_ptr,
    num_experts: tl.constexpr,
    block_size: tl.constexpr,
    numel: tl.constexpr,
    tokens_per_thread: tl.constexpr,
):
    pid = tl.program_id(0)
    start_idx = tl.load(cumsum_ptr + pid)
    end_idx = tl.load(cumsum_ptr + pid + 1)

    for i in range(start_idx, end_idx, block_size):
        tl.store(expert_ids_ptr + i // block_size, pid)

    start_idx = pid * tokens_per_thread
    off_t = pid * num_experts

    for i in range(start_idx, tl.minimum(start_idx + tokens_per_thread, numel)):
        expert_id = tl.load(topk_ids_ptr + i)
        token_cnt = tl.load(tokens_cnts_ptr + off_t + expert_id)
        rank_post_pad = token_cnt + tl.load(cumsum_ptr + expert_id)
        tl.store(sorted_token_ids_ptr + rank_post_pad, i)
        tl.store(tokens_cnts_ptr + off_t + expert_id, token_cnt + 1)


# Triton implementation based on:
# https://github.com/sgl-project/sglang/commit/ba5112ff691d791a9e38c6c71f59324a5fcb49d0
def moe_align_block_size_triton(
    topk_ids: torch.Tensor,
    num_experts: int,
    block_size: int,
    sorted_token_ids: torch.Tensor,
    expert_ids: torch.Tensor,
    num_tokens_post_pad: torch.Tensor,
    cumsum_buffer: torch.Tensor,
) -> None:
    numel = topk_ids.numel()
    grid = (num_experts,)
    tokens_cnts = torch.zeros(
        (num_experts + 1, num_experts), dtype=torch.int32, device=topk_ids.device
    )
    tokens_per_thread = ceil_div(numel, num_experts)

    moe_align_block_size_stage1[grid](
        topk_ids,
        tokens_cnts,
        num_experts,
        numel,
        tokens_per_thread,
    )
    moe_align_block_size_stage2[grid](
        tokens_cnts,
        num_experts,
    )
    moe_align_block_size_stage3[(1,)](
        num_tokens_post_pad,
        tokens_cnts,
        cumsum_buffer,
        num_experts,
        block_size,
    )
    moe_align_block_size_stage4[grid](
        topk_ids,
        sorted_token_ids,
        expert_ids,
        tokens_cnts,
        cumsum_buffer,
        num_experts,
        block_size,
        numel,
        tokens_per_thread,
    )


def moe_align_block_size_cuda(
    topk_ids: torch.Tensor,
    block_size: int,
    num_experts: int,
    expert_map: torch.Tensor = None,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Aligns the token distribution across experts to be compatible with block
    size for matrix multiplication.

    Parameters:
    - topk_ids: A tensor of shape [total_tokens, top_k] representing the
        top-k expert indices for each token.
    - block_size: The block size used in block matrix multiplication.
    - num_experts: The total number of experts.
    - expert_map: A tensor of shape [num_experts] that maps the expert index
        from the global space to the local index space of the current
        expert parallel shard. If the expert is not in the current expert
        parallel shard, the mapping is set to -1.

    Returns:
    - sorted_token_ids: A tensor containing the sorted token indices according
        to their allocated expert.
    - expert_ids: A tensor indicating the assigned expert index for each block.
    - num_tokens_post_padded: The total number of tokens after padding,
        ensuring divisibility by block_size.

    This function pads the number of tokens that each expert needs to process
    so that it is divisible by block_size.
    Padding ensures that during block matrix multiplication, the dimensions
    align correctly.

    Example:
    Given topk_ids = [[2, 3, 4], [1, 2, 4], [1, 3, 4], [1, 2, 3]],
    block_size = 4, and num_experts = 4:
    - We initially have 12 tokens (after repeating 'top_k' times) and 4 experts,
        with each expert needing to process 3 tokens.
    - As block_size is 4, we pad 1 token for each expert.
    - First, flatten topk_ids to [2, 3, 4, 1, 2, 4, 1, 3, 4, 1, 2, 3].
    - Then append padding tokens [12, 12, 12, 12] for each block.
    - After sorting by expert index, we obtain token_ids
        [3, 6, 9, 12, 0, 4, 10, 12, 1, 7, 11, 12, 2, 5, 8, 12].
        Tokens 12 are non-existent (padding) and are ignored in
        the subsequent matrix multiplication.
    - The padding ensures that the total number of tokens is now divisible
        by block_size for proper block matrix operations.
    """
    max_num_tokens_padded = topk_ids.numel() + num_experts * (block_size - 1)
    sorted_ids = torch.empty(
        (max_num_tokens_padded,), dtype=torch.int32, device=topk_ids.device
    )
    sorted_ids.fill_(topk_ids.numel())
    max_num_m_blocks = triton.cdiv(max_num_tokens_padded, block_size)
    # Expert ids must be zeroed out to prevent index out of bounds error while
    # mapping global expert ids to local expert ids in expert parallelism.
    expert_ids = torch.zeros(
        (max_num_m_blocks,), dtype=torch.int32, device=topk_ids.device
    )
    num_tokens_post_pad = torch.empty((1), dtype=torch.int32, device=topk_ids.device)
    token_cnts_buffer = torch.zeros(
        (num_experts + 1) * num_experts,
        dtype=torch.int32,
        device=topk_ids.device,
    )
    cumsum_buffer = torch.zeros(
        (num_experts + 1,), dtype=torch.int32, device=topk_ids.device
    )
    chitu_backend.cuda_moe_align_block_size(
        topk_ids,
        num_experts,
        block_size,
        sorted_ids,
        expert_ids,
        num_tokens_post_pad,
        token_cnts_buffer,
        cumsum_buffer,
    )
    if expert_map is not None:
        expert_ids = expert_map[expert_ids]

    return sorted_ids, expert_ids, num_tokens_post_pad


def moe_align_block_size_native(
    topk_ids: torch.Tensor,
    block_size: int,
    num_experts: int,
    expert_map: torch.Tensor = None,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Aligns the token distribution across experts to be compatible with block
    size for matrix multiplication.

    Parameters:
    - topk_ids: A tensor of shape [total_tokens, top_k] representing the
        top-k expert indices for each token.
    - block_size: The block size used in block matrix multiplication.
    - num_experts: The total number of experts.
    - expert_map: A tensor of shape [num_experts] that maps the expert index
        from the global space to the local index space of the current
        expert parallel shard. If the expert is not in the current expert
        parallel shard, the mapping is set to -1.

    Returns:
    - sorted_token_ids: A tensor containing the sorted token indices according
        to their allocated expert.
    - expert_ids: A tensor indicating the assigned expert index for each block.
    - num_tokens_post_padded: The total number of tokens after padding,
        ensuring divisibility by block_size.

    This function pads the number of tokens that each expert needs to process
    so that it is divisible by block_size.
    Padding ensures that during block matrix multiplication, the dimensions
    align correctly.

    Example:
    Given topk_ids = [[2, 3, 4], [1, 2, 4], [1, 3, 4], [1, 2, 3]],
    block_size = 4, and num_experts = 4:
    - We initially have 12 tokens (after repeating 'top_k' times) and 4 experts,
        with each expert needing to process 3 tokens.
    - As block_size is 4, we pad 1 token for each expert.
    - First, flatten topk_ids to [2, 3, 4, 1, 2, 4, 1, 3, 4, 1, 2, 3].
    - Then append padding tokens [12, 12, 12, 12] for each block.
    - After sorting by expert index, we obtain token_ids
        [3, 6, 9, 12, 0, 4, 10, 12, 1, 7, 11, 12, 2, 5, 8, 12].
        Tokens 12 are non-existent (padding) and are ignored in
        the subsequent matrix multiplication.
    - The padding ensures that the total number of tokens is now divisible
        by block_size for proper block matrix operations.
    """
    max_num_tokens_padded = topk_ids.numel() + num_experts * (block_size - 1)
    sorted_ids = torch.empty(
        (max_num_tokens_padded,), dtype=torch.int32, device=topk_ids.device
    )
    sorted_ids.fill_(topk_ids.numel())
    max_num_m_blocks = triton.cdiv(max_num_tokens_padded, block_size)
    # Expert ids must be zeroed out to prevent index out of bounds error while
    # mapping global expert ids to local expert ids in expert parallelism.
    expert_ids = torch.zeros(
        (max_num_m_blocks,), dtype=torch.int32, device=topk_ids.device
    )
    num_tokens_post_pad = torch.empty((1), dtype=torch.int32, device=topk_ids.device)
    cumsum_buffer = torch.zeros(
        (num_experts + 1,), dtype=torch.int32, device=topk_ids.device
    )
    moe_align_block_size_triton(
        topk_ids,
        num_experts,
        block_size,
        sorted_ids,
        expert_ids,
        num_tokens_post_pad,
        cumsum_buffer,
    )
    if expert_map is not None:
        expert_ids = expert_map[expert_ids]

    return sorted_ids, expert_ids, num_tokens_post_pad


def moe_align_block_size(
    topk_ids: torch.Tensor,
    block_size: int,
    num_experts: int,
    expert_map: torch.Tensor = None,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    if is_nvidia() or is_muxi():
        return moe_align_block_size_cuda(topk_ids, block_size, num_experts, expert_map)
    else:
        return moe_align_block_size_native(
            topk_ids, block_size, num_experts, expert_map
        )


@triton.jit
def _per_token_group_quant_fp8_colmajor(
    # Pointers to inputs and output
    y_ptr,
    y_q_ptr,
    y_s_ptr,
    group_size,
    # Num columns of y
    y_num_columns,
    y_row_stride,
    # Stride from one column to the next of y_s
    y_s_col_stride,
    # Avoid to divide zero
    eps,
    # Information for float8
    fp8_min,
    fp8_max,
    # Meta-parameters
    BLOCK: tl.constexpr,
):
    """A Triton-accelerated function to perform per-token-group
    quantization on a tensor.
    This function converts the tensor values into float8 values.
    """
    groups_per_row = y_num_columns // group_size

    # Map the program id to the row of X and Y it should compute.
    g_id = tl.program_id(0)
    row = g_id // groups_per_row
    row_g_id = g_id % groups_per_row

    y_ptr += (row * y_row_stride) + (row_g_id * group_size)
    y_q_ptr += g_id * group_size

    # Convert g_id the flattened block coordinate to 2D so we can index
    # into the output y_scales matrix
    blocks_per_row = y_num_columns // group_size
    scale_col = g_id % blocks_per_row
    scale_row = g_id // blocks_per_row
    y_s_ptr += scale_col * y_s_col_stride + scale_row

    cols = tl.arange(0, BLOCK)  # group_size <= BLOCK
    mask = cols < group_size

    y = tl.load(y_ptr + cols, mask=mask, other=0.0).to(tl.float32)
    # Quant
    _absmax = tl.maximum(tl.max(tl.abs(y)), eps)
    y_s = _absmax / fp8_max
    y_q = tl.clamp(y / y_s, fp8_min, fp8_max).to(y_q_ptr.dtype.element_ty)

    tl.store(y_q_ptr + cols, y_q, mask=mask)
    tl.store(y_s_ptr, y_s)


@triton.jit
def _per_token_group_quant_fp8(
    # Pointers to inputs and output
    y_ptr,
    y_q_ptr,
    y_s_ptr,
    group_size,
    # Num columns of y
    y_num_columns,
    y_row_stride,
    # Avoid to divide zero
    eps,
    # Information for float8
    fp8_min,
    fp8_max,
    # Meta-parameters
    BLOCK: tl.constexpr,
):
    """A Triton-accelerated function to perform per-token-group
    quantization on a tensor.
    This function converts the tensor values into float8 values.
    """
    groups_per_row = y_num_columns // group_size

    # Map the program id to the row of X and Y it should compute.
    g_id = tl.program_id(0)
    row = g_id // groups_per_row
    row_g_id = g_id % groups_per_row

    y_ptr += (row * y_row_stride) + (row_g_id * group_size)
    y_q_ptr += g_id * group_size
    y_s_ptr += g_id

    cols = tl.arange(0, BLOCK)  # N <= BLOCK
    mask = cols < group_size

    y = tl.load(y_ptr + cols, mask=mask, other=0.0).to(tl.float32)
    # Quant
    _absmax = tl.maximum(tl.max(tl.abs(y)), eps)
    y_s = _absmax / fp8_max
    y_q = tl.clamp(y / y_s, fp8_min, fp8_max).to(y_q_ptr.dtype.element_ty)

    tl.store(y_q_ptr + cols, y_q, mask=mask)
    tl.store(y_s_ptr, y_s)


def per_token_group_quant_fp8(
    x: torch.Tensor,
    group_size: int,
    eps: float = 1e-10,
    dtype: Optional[torch.dtype] = None,
    column_major_scales: bool = False,  # Changed, False
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Function to perform per-token-group quantization on an input tensor `x`.
    It converts the tensor values into signed float8 values and returns the
    quantized tensor along with the scaling factor used for quantization.
    Args:
        x: The input tensor with ndim >= 2.
        group_size: The group size used for quantization.
        eps: The minimum to avoid dividing zero.
        dtype: The dype of output tensor. Note that only `torch.float8_e4m3fn`
        is supported for now.
    Returns:
        Tuple[torch.Tensor, torch.Tensor]: The quantized tensor and the
        scaling factor for quantization.
    """
    if dtype is None:
        # dtype = (torch.float8_e4m3fnuz
        #          if current_platform.is_rocm() else torch.float8_e4m3fn)
        dtype = torch.float8_e4m3fn
    assert x.shape[-1] % group_size == 0, (
        f"the last dimension of `x` {x.shape[-1]} must be divisible "
        f"by `group_size` {group_size}"
    )
    assert x.stride(-1) == 1, "`x` groups must be contiguous"

    finfo = torch.finfo(dtype)
    fp8_min = finfo.min
    fp8_max = finfo.max

    x_q = torch.empty_like(x, device=x.device, dtype=dtype)
    M = x.numel() // group_size
    N = group_size
    if column_major_scales:
        shape = (x.shape[-1] // group_size,) + x.shape[:-1]
        x_s = torch.empty(shape, device=x.device, dtype=torch.float32).permute(-1, -2)
    else:
        shape = x.shape[:-1] + (x.shape[-1] // group_size,)
        x_s = torch.empty(shape, device=x.device, dtype=torch.float32)

    BLOCK = triton.next_power_of_2(N)
    # heuristics for number of warps
    num_warps = min(max(BLOCK // 256, 1), 8)
    num_stages = 1
    if column_major_scales:
        _per_token_group_quant_fp8_colmajor[(M,)](
            x,
            x_q,
            x_s,
            group_size,
            x.shape[1],
            x.stride(0),
            x_s.stride(1),
            eps,
            fp8_min=fp8_min,
            fp8_max=fp8_max,
            BLOCK=BLOCK,
            num_warps=num_warps,
            num_stages=num_stages,
        )
    else:
        _per_token_group_quant_fp8[(M,)](
            x,
            x_q,
            x_s,
            group_size,
            x.shape[1],
            x.stride(0),
            eps,
            fp8_min=fp8_min,
            fp8_max=fp8_max,
            BLOCK=BLOCK,
            num_warps=num_warps,
            num_stages=num_stages,
        )

    return x_q, x_s


def invoke_fused_moe_kernel(
    A: torch.Tensor,
    B: torch.Tensor,
    C: torch.Tensor,
    A_scale: Optional[torch.Tensor],
    B_scale: Optional[torch.Tensor],
    B_zp: Optional[torch.Tensor],
    topk_weights: torch.Tensor,
    topk_ids: torch.Tensor,
    sorted_token_ids: torch.Tensor,
    expert_ids: torch.Tensor,
    num_tokens_post_padded: torch.Tensor,
    mul_routed_weight: bool,
    top_k: int,
    config: Dict[str, Any],
    compute_type: tl.dtype,
    use_fp8_w8a8: bool,  # NOTE only support fp8_w8a8
    use_int8_w8a16: bool,
    use_int4_w4a16: bool,
    block_shape: Optional[List[int]] = None,
    soft_fp8: bool = False,
) -> None:
    assert topk_weights.stride(1) == 1
    assert sorted_token_ids.stride(0) == 1

    if use_fp8_w8a8:
        assert B_scale is not None
        if block_shape is None:
            # A, A_scale = ops.scaled_fp8_quant(A, A_scale)
            assert False
        elif not soft_fp8:
            assert len(block_shape) == 2
            block_n, block_k = block_shape[0], block_shape[1]
            A, A_scale = per_token_group_quant_fp8(A, block_k)
            assert triton.cdiv(A.shape[-1], block_k) == A_scale.shape[-1]
            assert triton.cdiv(B.shape[-2], block_n) == B_scale.shape[-2]
            assert triton.cdiv(B.shape[-1], block_k) == B_scale.shape[-1]
        else:
            B = B.view(dtype=torch.uint8)
            assert A_scale is None
            assert B_scale is not None
    elif use_int8_w8a16 or use_int4_w4a16:
        assert B_scale is not None
        assert block_shape is None or block_shape[0] == 0
    else:
        assert A_scale is None
        assert B_scale is None

    # Some of our platforms only has Triton with low versions, where these is no `tl.cast`
    # which is used for initializing a constant with a given type. Therefore, we need to
    # pass `fp8_to_fp32_scale` as a constant from outside.
    fp8_to_fp32_scale = struct.unpack(">f", bytes.fromhex("7b800000"))[0]

    EM = sorted_token_ids.shape[0]
    if A.shape[0] < config["BLOCK_SIZE_M"]:
        # optimize for small batch_size.
        # We assume that top_ids of each token is unique, so
        # so num_valid_experts <= batch_size <= BLOCK_SIZE_M,
        # and we can skip some invalid blocks.
        EM = min(sorted_token_ids.shape[0], A.shape[0] * top_k * config["BLOCK_SIZE_M"])
    grid = lambda META: (
        triton.cdiv(EM, META["BLOCK_SIZE_M"])
        * triton.cdiv(B.shape[1], META["BLOCK_SIZE_N"]),
    )

    fused_moe_kernel[grid](
        A,
        B,
        C,
        A_scale,
        B_scale,
        topk_weights,
        sorted_token_ids,
        expert_ids,
        num_tokens_post_padded,
        B.shape[1],
        A.shape[1],
        EM,
        topk_ids.numel(),
        A.stride(0),
        A.stride(1),
        B.stride(0),
        B.stride(2),
        B.stride(1),
        C.stride(1),
        C.stride(2),
        A_scale.stride(0) if A_scale is not None and A_scale.ndim == 2 else 0,
        A_scale.stride(1) if A_scale is not None and A_scale.ndim == 2 else 0,
        B_scale.stride(0) if B_scale is not None and B_scale.ndim >= 2 else 0,
        B_scale.stride(2) if B_scale is not None and B_scale.ndim == 3 else 0,
        B_scale.stride(1) if B_scale is not None and B_scale.ndim >= 2 else 0,
        0 if block_shape is None else block_shape[0],
        0 if block_shape is None else block_shape[1],
        MUL_ROUTED_WEIGHT=mul_routed_weight,
        top_k=top_k,
        compute_type=compute_type,
        use_fp8_w8a8=use_fp8_w8a8,
        use_int8_w8a16=use_int8_w8a16,
        soft_fp8=soft_fp8,
        fp8_to_fp32_scale=fp8_to_fp32_scale,
        **config,
    )


def get_default_config(
    M: int,
    E: int,
    N: int,
    K: int,
    topk: int,
    dtype: Optional[str],
    is_marlin: bool,
    block_shape: Optional[List[int]] = None,
) -> Dict[str, int]:
    if dtype == "fp8_w8a8" and block_shape is not None:
        # Block-wise quant: BLOCK_SIZE_N must be divisible by block_shape[0]
        # BLOCK_SIZE_K must be divisible by block_shape[1]
        config = {
            "BLOCK_SIZE_M": 64,
            "BLOCK_SIZE_N": block_shape[0],
            "BLOCK_SIZE_K": block_shape[1],
            "GROUP_SIZE_M": 32,
            "num_warps": 4,
            "num_stages": 3,
        }
    else:
        config = {
            "BLOCK_SIZE_M": 64,
            "BLOCK_SIZE_N": 64,
            "BLOCK_SIZE_K": 32,
            "GROUP_SIZE_M": 8,
        }
        # A heuristic: fused marlin works faster with this config for small M
        if M <= E or (is_marlin and M <= 32):
            config = {
                "BLOCK_SIZE_M": 16,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 64,
                "GROUP_SIZE_M": 1,
            }
    return config


def try_get_optimal_moe_config(
    w1_shape: Tuple[int, ...],
    w2_shape: Tuple[int, ...],
    top_k: int,
    dtype: Optional[str],
    M: int,
    is_marlin: bool = False,
    block_shape: Optional[List[int]] = None,
):
    # First try to load optimal config from the file
    E, _, N = w2_shape

    config = get_default_config(
        M, E, N, w1_shape[2], top_k, dtype, is_marlin, block_shape
    )
    return config


def get_config_dtype_str(
    dtype: torch.dtype,
    use_int4_w4a16: Optional[bool] = False,
    use_int8_w8a16: Optional[bool] = False,
    use_fp8_w8a8: Optional[bool] = False,
):
    if use_fp8_w8a8:
        return "fp8_w8a8"
    elif use_int8_w8a16:
        return "int8_w8a16"
    elif use_int4_w4a16:
        return "int4_w8a16"
    elif dtype == torch.float:
        # avoiding cases where kernel fails when float32 MoE
        # use fp16/bfloat16 configs
        return "float32"
    return None


def inplace_fused_experts(
    hidden_states: torch.Tensor,
    w1: torch.Tensor,
    w2: torch.Tensor,
    topk_weights: torch.Tensor,
    topk_ids: torch.Tensor,
    activation: str = "silu",
    use_fp8_w8a8: bool = False,
    use_int8_w8a16: bool = False,
    use_int4_w4a16: bool = False,
    global_num_experts: int = -1,
    expert_map: Optional[torch.Tensor] = None,
    w1_scale: Optional[torch.Tensor] = None,
    w2_scale: Optional[torch.Tensor] = None,
    w1_zp: Optional[torch.Tensor] = None,
    w2_zp: Optional[torch.Tensor] = None,
    a1_scale: Optional[torch.Tensor] = None,
    a2_scale: Optional[torch.Tensor] = None,
    block_shape: Optional[List[int]] = None,
    soft_fp8: bool = False,
) -> None:
    fused_experts_impl(
        hidden_states,
        w1,
        w2,
        topk_weights,
        topk_ids,
        True,
        activation,
        use_fp8_w8a8,
        use_int8_w8a16,
        use_int4_w4a16,
        global_num_experts,
        expert_map,
        w1_scale,
        w2_scale,
        w1_zp,
        w2_zp,
        a1_scale,
        a2_scale,
        block_shape,
        soft_fp8=soft_fp8,
    )


def outplace_fused_experts(
    hidden_states: torch.Tensor,
    w1: torch.Tensor,
    w2: torch.Tensor,
    topk_weights: torch.Tensor,
    topk_ids: torch.Tensor,
    activation: str = "silu",
    use_fp8_w8a8: bool = False,
    use_int8_w8a16: bool = False,
    use_int4_w4a16: bool = False,
    global_num_experts: int = -1,
    expert_map: Optional[torch.Tensor] = None,
    w1_scale: Optional[torch.Tensor] = None,
    w2_scale: Optional[torch.Tensor] = None,
    w1_zp: Optional[torch.Tensor] = None,
    w2_zp: Optional[torch.Tensor] = None,
    a1_scale: Optional[torch.Tensor] = None,
    a2_scale: Optional[torch.Tensor] = None,
    block_shape: Optional[List[int]] = None,
    soft_fp8: bool = False,
) -> torch.Tensor:
    return fused_experts_impl(
        hidden_states,
        w1,
        w2,
        topk_weights,
        topk_ids,
        False,
        activation,
        use_fp8_w8a8,
        use_int8_w8a16,
        use_int4_w4a16,
        global_num_experts,
        expert_map,
        w1_scale,
        w2_scale,
        w1_zp,
        w2_zp,
        a1_scale,
        a2_scale,
        block_shape,
        soft_fp8=soft_fp8,
    )


def fused_experts(
    hidden_states: torch.Tensor,
    w1: torch.Tensor,
    w2: torch.Tensor,
    topk_weights: torch.Tensor,
    topk_ids: torch.Tensor,
    inplace: bool = False,
    activation: str = "silu",
    use_fp8_w8a8: bool = False,
    use_int8_w8a16: bool = False,
    use_int4_w4a16: bool = False,
    global_num_experts: int = -1,
    expert_map: Optional[torch.Tensor] = None,
    w1_scale: Optional[torch.Tensor] = None,
    w2_scale: Optional[torch.Tensor] = None,
    w1_zp: Optional[torch.Tensor] = None,
    w2_zp: Optional[torch.Tensor] = None,
    a1_scale: Optional[torch.Tensor] = None,
    a2_scale: Optional[torch.Tensor] = None,
    block_shape: Optional[List[int]] = None,
    soft_fp8: bool = False,
) -> torch.Tensor:

    if inplace:
        inplace_fused_experts(
            hidden_states,
            w1,
            w2,
            topk_weights,
            topk_ids,
            activation,
            use_fp8_w8a8,
            use_int8_w8a16,
            use_int4_w4a16,
            global_num_experts,
            expert_map,
            w1_scale,
            w2_scale,
            w1_zp,
            w2_zp,
            a1_scale,
            a2_scale,
            block_shape,
            soft_fp8=soft_fp8,
        )
        return hidden_states
    else:
        return torch.ops.vllm.outplace_fused_experts(
            hidden_states,
            w1,
            w2,
            topk_weights,
            topk_ids,
            activation,
            use_fp8_w8a8,
            use_int8_w8a16,
            use_int4_w4a16,
            global_num_experts,
            expert_map,
            w1_scale,
            w2_scale,
            w1_zp,
            w2_zp,
            a1_scale,
            a2_scale,
            block_shape,
            soft_fp8=soft_fp8,
        )


def fused_experts_impl(
    hidden_states: torch.Tensor,
    w1: torch.Tensor,
    w2: torch.Tensor,
    topk_weights: torch.Tensor,
    topk_ids: torch.Tensor,
    inplace: bool = False,
    activation: str = "silu",
    use_fp8_w8a8: bool = False,
    use_int8_w8a16: bool = False,
    use_int4_w4a16: bool = False,
    global_num_experts: int = -1,
    expert_map: Optional[torch.Tensor] = None,
    w1_scale: Optional[torch.Tensor] = None,
    w2_scale: Optional[torch.Tensor] = None,
    w1_zp: Optional[torch.Tensor] = None,
    w2_zp: Optional[torch.Tensor] = None,
    a1_scale: Optional[torch.Tensor] = None,
    a2_scale: Optional[torch.Tensor] = None,
    block_shape: Optional[List[int]] = None,
    soft_fp8: bool = False,
):
    # Check constraints.
    if use_int4_w4a16:
        assert hidden_states.shape[1] // 2 == w1.shape[2], "Hidden size mismatch"
    else:
        assert hidden_states.shape[1] == w1.shape[2], "Hidden size mismatch"

    assert topk_weights.shape == topk_ids.shape, "topk shape mismatch"
    assert hidden_states.is_contiguous(), "Hidden_states must be contiguous"
    assert w1.is_contiguous(), "Expert weights1 must be contiguous"
    assert w2.is_contiguous(), "Expert weights2 must be contiguous"
    assert hidden_states.dtype in [torch.float32, torch.float16, torch.bfloat16]

    num_tokens, _ = hidden_states.shape
    E, N, _ = w1.shape
    if global_num_experts == -1:
        global_num_experts = E
    top_k_num = topk_ids.shape[1]
    # We execute the fused_moe kernel in chunks to circumvent this issue:
    # https://github.com/vllm-project/vllm/issues/5938
    # CHUNK_SIZE = envs.VLLM_FUSED_MOE_CHUNK_SIZE
    CHUNK_SIZE = 32768
    M = min(num_tokens, CHUNK_SIZE)
    config_dtype = get_config_dtype_str(
        use_fp8_w8a8=use_fp8_w8a8,
        use_int8_w8a16=use_int8_w8a16,
        use_int4_w4a16=use_int4_w4a16,
        dtype=hidden_states.dtype,
    )

    get_config_func = functools.partial(
        try_get_optimal_moe_config,
        w1.shape,
        w2.shape,
        top_k_num,
        config_dtype,
        block_shape=block_shape,
    )

    config = get_config_func(M)

    intermediate_cache1 = torch.empty(
        (M, top_k_num, N), device=hidden_states.device, dtype=hidden_states.dtype
    )
    intermediate_cache2 = torch.empty(
        (M * top_k_num, N // 2), device=hidden_states.device, dtype=hidden_states.dtype
    )
    intermediate_cache3 = torch.empty(
        (M, top_k_num, w2.shape[1]),
        device=hidden_states.device,
        dtype=hidden_states.dtype,
    )

    if hidden_states.dtype == torch.bfloat16:
        compute_type = tl.bfloat16
    elif hidden_states.dtype == torch.float16:
        compute_type = tl.float16
    elif hidden_states.dtype == torch.float32:
        compute_type = tl.float32
    else:
        raise ValueError(f"Unsupported compute_type: {hidden_states.dtype}")

    if inplace:
        out_hidden_states = hidden_states
    else:
        out_hidden_states = torch.empty_like(hidden_states)

    for chunk in range((num_tokens // CHUNK_SIZE) + 1):
        begin_chunk_idx, end_chunk_idx = (
            chunk * CHUNK_SIZE,
            min((chunk + 1) * CHUNK_SIZE, num_tokens),
        )
        curr_hidden_states = hidden_states[begin_chunk_idx:end_chunk_idx]
        tokens_in_chunk, _ = curr_hidden_states.shape

        if tokens_in_chunk == 0:
            break

        if tokens_in_chunk < CHUNK_SIZE and chunk > 0:
            # Adjust the intermediate cache size and config for the last
            # chunk. Note that in most cases we only have one chunk
            # so the cache size and config are already set correctly and
            # do not need to be adjusted.
            intermediate_cache1 = intermediate_cache1[:tokens_in_chunk]
            intermediate_cache2 = intermediate_cache2[
                : tokens_in_chunk * topk_ids.shape[1]
            ]
            intermediate_cache3 = intermediate_cache3[:tokens_in_chunk]
            config = get_config_func(tokens_in_chunk)

        curr_topk_ids = topk_ids[begin_chunk_idx:end_chunk_idx]
        curr_topk_weights = topk_weights[begin_chunk_idx:end_chunk_idx]

        sorted_token_ids, expert_ids, num_tokens_post_padded = moe_align_block_size(
            curr_topk_ids, config["BLOCK_SIZE_M"], global_num_experts, expert_map
        )

        invoke_fused_moe_kernel(
            curr_hidden_states,
            w1,
            intermediate_cache1,
            a1_scale,
            w1_scale,
            w1_zp,
            curr_topk_weights,
            curr_topk_ids,
            sorted_token_ids,
            expert_ids,
            num_tokens_post_padded,
            False,
            top_k_num,
            config,
            compute_type=compute_type,
            use_fp8_w8a8=use_fp8_w8a8,
            use_int8_w8a16=use_int8_w8a16,
            use_int4_w4a16=use_int4_w4a16,
            block_shape=block_shape,
            soft_fp8=soft_fp8,
        )

        if activation == "silu":
            intermediate_cache2 = SiluAndMul()(intermediate_cache1.view(-1, N))
        else:
            raise ValueError(f"Unsupported FusedMoe activation: {activation}")

        invoke_fused_moe_kernel(
            intermediate_cache2,
            w2,
            intermediate_cache3,
            a2_scale,
            w2_scale,
            w2_zp,
            curr_topk_weights,
            curr_topk_ids,
            sorted_token_ids,
            expert_ids,
            num_tokens_post_padded,
            True,
            1,
            config,
            compute_type=compute_type,
            use_fp8_w8a8=use_fp8_w8a8,
            use_int8_w8a16=use_int8_w8a16,
            use_int4_w4a16=use_int4_w4a16,
            block_shape=block_shape,
            soft_fp8=soft_fp8,
        )

        def moe_sum(input: torch.Tensor, output: torch.Tensor):
            output.copy_(input.sum(dim=1))

        moe_sum(
            intermediate_cache3.view(*intermediate_cache3.shape),
            out_hidden_states[begin_chunk_idx:end_chunk_idx],
        )

    return out_hidden_states



import os
import sys

from setuptools import Extension, setup, find_packages
from setuptools.command.build_ext import build_ext
from setuptools.command.build_py import build_py
from Cython.Build import cythonize
from torch.utils.cpp_extension import BuildExtension, CUDAExtension

setup_dir = os.path.dirname(os.path.abspath(__file__))

ext_modules = [
    # We use CUDAExtension instead of CMake for native sources, because many of the non-NVIDIA GPUs have
    # their custom CUDAExtension, but not their custom CMake support.
    CUDAExtension(
        name="chitu_backend",
        sources=[
            "./csrc/binding.cpp",
            "./csrc/moe_align_kernel.cu",
            "./csrc/fused_shared_experts_kernel.cu",
        ],
        extra_compile_args={
            "cxx": ["-std=c++17"],
            "nvcc": ["-std=c++17"],
        },
        include_dirs=[os.path.join(setup_dir, "third_party/spdlog/include")],
    )
]


cython_unsafe_files = [
    "triton_kernels.py",
    "fused_moe.py",
    "triton_decode_attention.py",
]

def is_cython_unsafe(path):
    for unsafe_file in cython_unsafe_files:
        if str(path).endswith(unsafe_file):
            return True
    return False

def find_py_modules(directory):
    modules = []
    for root, _, files in os.walk(directory):
        for file in files:
            if file.endswith(".py"):
                if not is_cython_unsafe(os.path.join(root, file)):
                    module_name = os.path.splitext(os.path.join(root, file))[0].replace(os.sep, ".")
                    modules.append(module_name)
    return modules

def create_cython_extensions(directory):
    extensions = []
    for module in find_py_modules(directory):
        extension = Extension(module, [module.replace(".", os.sep) + ".py"])
        extensions.append(extension)
    return extensions

class SkipBuildPy(build_py):
    def find_package_modules(self, package, package_dir):
        modules = super().find_package_modules(package, package_dir)
        filtered_modules = [
            (pkg, mod, file) for (pkg, mod, file) in modules if is_cython_unsafe(file)
        ]
        return filtered_modules

my_build_py = build_py
if os.environ.get("CINFER_WITH_CYTHON", "0") != "0":
    ext_modules += cythonize(create_cython_extensions("chitu"))
    my_build_py = SkipBuildPy

# The information here can also be placed in setup.cfg - better separation of
# logic and declaration, and simpler if you include description/version in a file.
setup(
    name="chitu",
    version="0.1.0",
    install_requires=[
        # Don't put `torch` here because it requires downloading from a specific source
        "transformers",
        "fire",
        "tiktoken>=0.7.0",  # Required by glm4
        "blobfile",
        "faker",
        "hydra-core",
        "fastapi",
        "uvicorn",
        "tqdm",
        "accelerate",
        "einops",
        "typing-extensions",
    ],
    extras_require={
        "quant": [
            "optimum",
            "bitsandbytes",
            "EETQ @ file://localhost" + os.path.join(setup_dir, "third_party/EETQ"),
            "awq_inference_engine @ file://localhost"
            + os.path.join(setup_dir, "third_party/llm-awq/awq/kernels"),
            "auto_gptq @ file://localhost"
            + os.path.join(setup_dir, "third_party/AutoGPTQ"),
            "w8a8gemm @ file://localhost"
            + os.path.join(setup_dir, "third_party/nv_w8a8_kernels/w8a8gemm"),
            "w8a8gemv @ file://localhost"
            + os.path.join(setup_dir, "third_party/nv_w8a8_kernels/w8a8gemv"),
        ],
        "muxi_layout_kernels": [
            "muxi_layout_kernels @ file://localhost"
            + os.path.join(setup_dir, "third_party/muxi_layout_kernels"),
        ],
        "muxi_w8a8_kernels": [
            "tbsgemm @ file://localhost"
            + os.path.join(setup_dir, "third_party/muxi_w8a8_kernels/w8a8"),
        ],
        "flash_attn": [
            "flash-attn",
            # Although `flash-attn` is available in PyPI, don't make it a required
            # dependency, because its installation runs forever on some platforms.
        ],
        "flashinfer": [
            "flashinfer-python",
        ],
        "flash_mla": [
            "flash_mla @ file://localhost"
            + os.path.join(setup_dir, "third_party/FlashMLA"),
        ],
    },
    packages=find_packages(),
    ext_modules=ext_modules,
    cmdclass={"build_ext": BuildExtension, "build_py": my_build_py},
    package_data={"chitu": ["config/**/*.yaml"]},
)
